"""Sample class which carries out suppression operations based on the location of 
the frontier of the fire.

Note: This example is not intended to be a robust firefighting algorithm, it is to be used 
as a reference on how agents can be used in conjuction with the fire interface to formulate
a response to a fire.

To run this example code, start a fire sim and select this fiel as the "User Module"
"""

import numpy as np
from sklearn.cluster import KMeans
import heapq

from base_classes.control_base import ControlClass
from fire_simulator.fire import FireSim
from v_sample_agent import SampleAgent
from utilities.fire_util import FuelConstants as fc

class SuppressionExample(ControlClass):
    def __init__(self, fire:FireSim):
        # Define time agents should dispatch
        self.operation_start_time_h = 0.33

        # Initialize empty list to store agents 
        self.agents = []

        # Define number of agents
        num_agents = 10

        # Randomly place agents
        for i in range(num_agents):
            a_x = np.random.random() * fire.x_lim
            a_y = np.random.random() * fire.y_lim

            a = SampleAgent(i, a_x, a_y, fire, label=f"{i}")
            a.max_tasks = np.inf

            fire.add_agent(a)
            self.agents.append(a)

        self.soak_complete = False

    # process_state is called by the FireSim every iteration
    def process_state(self, fire: FireSim) -> None:
        # Check if it is time to dispatch agents
        if fire.curr_time_h > self.operation_start_time_h:
            if all(len(a.task_list) == 0 for a in self.agents):
                # Create tasks
                clusters, means = self.create_soak_tasks(fire)
                
                # Assign tasks
                self.assign_soak_tasks(clusters, means)

                self.soak_complete = True

        # Step each agent forward one time step
        self.iterate_agents()

    def create_soak_tasks(self, fire: FireSim) -> list:
        tot_x = 0
        tot_y = 0

        # Get the fire frontier sum up x and y coordinates
        frontier_points = []
        for cell in fire.frontier:
            frontier_points.append((cell.x_pos, cell.y_pos))
            tot_x += cell.x_pos
            tot_y += cell.y_pos

        # Calculate the average fire frontier x and y coordinates
        avg_frontier_x = tot_x / len(frontier_points)
        avg_frontier_y = tot_y / len(frontier_points)

        # Generate vectors from the average frontier coordinate to each frontier point
        frontier_vecs = [[p[0] - avg_frontier_x, p[1] - avg_frontier_y] for p in frontier_points]

        # Generate task points by projecting out 300 meters from each frontier point
        task_points = []
        for i, p in enumerate(frontier_points):
            unit_proj_vec = frontier_vecs[i]/np.linalg.norm(frontier_vecs[i])
            task_points.append([p[0] + 300 * unit_proj_vec[0], p[1] + 300 * unit_proj_vec[1]])

        # group task points into a group for each agent
        num_groups = np.min([len(self.agents), len(task_points)])

        # Use kmeans to cluster the task points
        kmeans = KMeans(n_clusters=num_groups, n_init=10)
        kmeans.fit(task_points)

        clusters = {}
        for i, label in enumerate(kmeans.labels_):
            if label not in clusters:
                clusters[label] = []
            clusters[label].append(task_points[i])

        means = kmeans.cluster_centers_.tolist()

        return clusters, means

    def assign_soak_tasks(self, clusters, means):
        # A boolean list to track assigned clusters
        assigned_tasks = [False] * len(clusters)
        
        # Iterate through agents and calculate the best cluster for each agent based on distance
        for agent in self.agents:
            min_dist = np.inf
            best_cluster = None

            for i, mean in enumerate(means):
                if not assigned_tasks[i]:  # Check if this cluster has been assigned
                    dist = np.sqrt((agent.x - mean[0])**2 + (agent.y - mean[1])**2)

                    if dist < min_dist:
                        min_dist = dist  # Update min_dist
                        best_cluster = i

            task_list = clusters[best_cluster]

            # Append tasks to selected agent's task list
            for task in task_list:
                agent.task_list.append((task, 'soak'))

            assigned_tasks[best_cluster] = True  # Mark this cluster as assigned

    # Call the complete_tasks function for each agent each time-step
    def iterate_agents(self):
        for agent in self.agents:
            agent.complete_tasks()







"""Module used to run the application that allows users to generate a new map file.
"""

from typing import Tuple
import xml.etree.ElementTree as ET
import json
import pickle
import sys
import os
import rasterio
import requests
import pyproj
import utm
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, BoundaryNorm
from shapely.geometry import Polygon, LineString
from shapely.geometry import mapping
from scipy.interpolate import RectBivariateSpline
from scipy import ndimage, stats
import numpy as np

from .utilities.file_io import MapGenFileSelector
from .utilities.map_drawer import PolygonDrawer
from .utilities.fire_util import RoadConstants as rc
from .utilities.fire_util import FuelConstants as fc

DATA_RESOLUTION = 30 # meters

def generate_map_from_file(file_params: dict, data_res: float, min_cell_size: float):
    """Generate a simulation map. Take in user's selections of fuel and topography files
    along with the drawings they overlay for initial ignitions and fire-breaks and generate a
    usable map.

    :param file_params: Dictionary containing the input parameters/files to generate map from
    :type file_parmas: dict
    :param data_res: Resolution of the fuel and elevation data
    :type data_res: float
    :param min_cell_size: minimum cell size for this map, used for interpolation of elevation
                          and fuel data
    :type min_cell_size: float
    :raises ValueError: if elevation and fuel data selected are not from the same region
    """
    # Initialize figure for GUI
    fig = plt.figure(figsize=(15, 10))
    plt.tick_params(left = False, right = False, bottom = False, labelleft = False,
                    labelbottom = False)

    # Get file paths for all data files
    save_path = file_params['Output Map Folder']
    top_path  = file_params['Topography Map Path']
    fuel_path = file_params['Fuel Map Path']

    uniform_fuel = file_params["Uniform Fuel"]
    uniform_elev = file_params["Uniform Elev"]

    # Get user choice for importing roads
    import_roads = file_params['Import Roads']

    if not uniform_elev and not uniform_fuel:
        fuel_data, fuel_bounds = parse_fuel_data(fuel_path, import_roads)
        topography_data, elev_bounds = parse_elevation_data(top_path, data_res, min_cell_size)

        # Ensure that the elevation and fuel data are from same region
        for e_bound, f_bound in zip(elev_bounds, fuel_bounds):
            if np.abs(e_bound - f_bound) > 1:
                raise ValueError('The elevation and fuel data are not from the same region')

        bounds = elev_bounds

        if fuel_data['width_m'] != topography_data['width_m'] or fuel_data['height_m'] != topography_data['height_m']:
            # set bounds to the smaller of the 2

            width_m = np.min([fuel_data['width_m'], topography_data['width_m']])
            height_m = np.min([fuel_data['height_m'], topography_data['height_m']])

            fuel_data['width_m'] = int(width_m)
            topography_data['width_m'] = int(width_m)
            fuel_data['height_m'] = int(height_m)
            topography_data['height_m'] = int(height_m)

    elif uniform_fuel and not uniform_elev:
        # parse elevation map first
        topography_data, elev_bounds = parse_elevation_data(top_path, data_res, min_cell_size)
        bounds = elev_bounds

        # create uniform fuel map based on dimensions
        rows, cols = topography_data["rows"], topography_data["cols"]
        fuel_type = file_params["Fuel Type"]
        fuel_data = create_uniform_fuel_map(rows, cols, fuel_type)

    elif not uniform_fuel and uniform_elev:
        # parse fuel map first
        fuel_data, fuel_bounds = parse_fuel_data(fuel_path, import_roads)
        bounds = fuel_bounds

        # create uniform elevation map based on dimensions
        rows, cols = fuel_data["height_m"], fuel_data["width_m"]
        topography_data = create_uniform_elev_map(rows, cols)
    else:
        # get sim size from parameters
        rows, cols = file_params["height m"], file_params["width m"]
        fuel_type = file_params["Fuel Type"]

        # create uniform elevation and fuel maps based on sim size
        topography_data = create_uniform_elev_map(rows, cols)
        fuel_data = create_uniform_fuel_map(rows, cols, fuel_type)
        bounds = None

    if import_roads:
        # get road data
        metadata_path = file_params['Metadata Path']
        road_data, bounds = get_road_data(metadata_path)
        if road_data is not None:
            roads = parse_road_data(road_data, bounds, fuel_data)
    else:
        roads = None

    # Get user input from GUI
    user_data = get_user_data(fig)

    save_to_file(save_path, fuel_data, topography_data, roads, user_data, bounds)

def save_to_file(save_path: str, fuel_data: dict, topography_data: dict,
                roads: list, user_data: dict, bounds: list):
    """Save all generated map data to a file specified in 'save_path'

    :param save_path: Path to save the map files in
    :type save_path: str
    :param fuel_data: Dictionary containing all data relevant to the fuel map
    :type fuel_data: dict
    :param topography_data: Dictionary containing all data relevant to the topography map
    :type topography_data: dict
    :param roads: list of points considered roads each element formatted as ((x, y), road_type)
    :type roads: list
    :param user_data: Dictionary containing the data generated by user in the drawing process
    :type user_data: dict
    :param bounds: list containing the geographic bounds of the map, corner coordinates listed in
    following order:[south, north, west, east]
    :type bounds: list
    """
    data = {}

    # Save fuel data
    fuel_path = save_path + '/fuel.npy'

    np.save(fuel_path, fuel_data['map'])

    if bounds is None:
        data['geo_bounds'] = None

    else:
        data['geo_bounds'] = {
            'south': bounds[0],
            'north': bounds[1],
            'west': bounds[2],
            'east': bounds[3] 
        }

    data['fuel'] = {'file': fuel_path,
                        'width_m': fuel_data['width_m'],
                        'height_m': fuel_data['height_m'],
                        'rows': fuel_data['rows'],
                        'cols': fuel_data['cols'],
                        'resolution': fuel_data['resolution'],
                        'uniform': fuel_data['uniform']
                    }
    
    if fuel_data['uniform']:
        data['fuel']['fuel type'] = fuel_data['fuel type']


    # Save topography data
    topography_path = save_path + '/topography.npy'
    np.save(topography_path, topography_data['map'])

    data['topography'] = {'file': topography_path,
                        'width_m': topography_data['width_m'],
                        'height_m': topography_data['height_m'],
                        'rows': topography_data['rows'],
                        'cols': topography_data['cols'],
                        'resolution': topography_data['resolution'],
                        'uniform': topography_data['uniform']
                        }

    # Save the roads data
    road_path = save_path + '/roads.pkl'
    with open(road_path, 'wb') as f:
        pickle.dump(roads, f)

    data['roads'] = {'file': road_path}

    data = data | user_data

    # Save data to JSON
    folder_name = os.path.basename(save_path)
    with open(save_path + "/" + folder_name + ".json", 'w') as f:
        json.dump(data, f, indent=4)

def get_road_data(path:str) -> Tuple[dict, list]:
    """Function that queries the openStreetMap API for the road data at the same region as the
    elevation and fuel maps
    
    :param path: path to the metadata file used to find the geographic region to pull from OSM API
    :type path: str
    :return: tuple with dictionary of road data and a list of the geobounds
    :rtype: Tuple[dict, list]
    """
    # Load the xml file
    tree = ET.parse(path)
    root = tree.getroot()

    west_bounding = float(root.find(".//westbc").text)
    east_bounding = float(root.find(".//eastbc").text)
    north_bounding = float(root.find(".//northbc").text)
    south_bounding = float(root.find(".//southbc").text)

    overpass_url = "http://overpass-api.de/api/interpreter"

    overpass_query = f"""
    [out:json];
    (way["highway"]
    ({south_bounding}, {west_bounding}, {north_bounding}, {east_bounding});
    );
    out body;
    >;
    out skel qt;
    """

    print(f"Querying OSM for road data at: [west: {west_bounding}, north: {north_bounding}," +
          f"east: {east_bounding}, south: {south_bounding}]")

    print("Awaiting response...")

    response = requests.get(overpass_url, params={'data': overpass_query})

    if response.status_code != 200:
        print(f"WARNING: Request failed with status {response.status_code}," +
              "proceeding without road data")

        print(response.text)
        return None

    road_data = response.json()

    print("Road data retrieved successfully!")

    return road_data, [south_bounding, north_bounding, west_bounding, east_bounding]

def parse_road_data(road_data: dict, bounds: list, data: dict) -> list:
    """Function that takes the raw road data, trims it to fit with the map and interpolates it to
    decrease the spacing between points on roads.

    :param road_data: dictionary containing raw road data from openStreetMap
    :type road_data: dict
    :param bounds: geographic bounds of the region the data is from
    :type bounds: list
    :param data: dictionary containing data about the fuel or elevation map
    :type data: dict
    :return: list of roads in the form of [((x,y), road type)]
    :rtype: list
    """
    # Extract the bounding coordinates from the data dictionary
    bbox = {
        'south': bounds[0],
        'north': bounds[1],
        'west': bounds[2],
        'east': bounds[3] 
    }

    # Calculate the central point of the bounding box
    central_lat = (bbox['south'] + bbox['north']) / 2
    central_lon = (bbox['west'] + bbox['east']) / 2

    # Get the UTM zone of the central point
    _, _, utm_zone, _ = utm.from_latlon(central_lat, central_lon)

    # Get projection based on the utm_zone
    proj = pyproj.Proj(proj='utm', zone=utm_zone, ellps='WGS84')

    # A dict to hold node coordinates
    nodes = {}

    # Get the origin in x,y coordinates
    origin_x, origin_y = proj(bbox['west'], bbox['south'])

    # First pass: get all node elements with their coordinates
    for element in road_data['elements']:
        if element['type'] == 'node':
            node_id = element['id']
            lat = element['lat']
            lon = element['lon']
            x, y = proj(lon, lat)

            # account for buffer on other data
            x -= 120
            y -= 120

            nodes[node_id] = (x - origin_x, y - origin_y)

    # Second pass: get all way elements that represent major roads
    roads = []
    for element in road_data['elements']:
        if element['type'] == 'way' and 'highway' in element['tags']:
            road_type = element['tags']['highway']
            node_ids = element['nodes']

            if road_type in rc.major_road_types:
                road = [nodes[node_id] for node_id in node_ids if node_id in nodes]
                if road:
                    roads.append((road, road_type))

    # Interpolate points so that each point in roads is at most 0.5m apart
    roads = interpolate_points(roads, 0.5)

    for road in roads:
        x_trimmed = []
        y_trimmed = []

        x, y = zip(*road[0])
        for i in range(len(x)):
            if 0 < x[i]/30 < data['cols']-1 and 0 < y[i]/30 < data['rows']-1:
                x_trimmed.append(x[i]/30)
                y_trimmed.append(y[i]/30)

        x = tuple(xi for xi in x_trimmed)
        y = tuple(yi for yi in y_trimmed)

        plt.plot(x, y, color=rc.road_color_mapping[road[1]])

    return roads

def interpolate_points(roads: list, max_spacing_m: float) -> list:
    """Interpolate points along a road so that every consecutive pair of points is less than a
    certain distance apart

    :param roads: list containing road data in the form [((x,y), road type)]
    :type roads: list
    :param max_spacing_m: maximum distance in meters two consecutive points can be apart
    :type max_spacing_m: float
    :return: list in the same form as 'roads' but with new interpolated points added in
    :rtype: list
    """
    interpolated_roads = []
    for road in roads:
        interpolated_road = []
        for i in range(len(road[0]) - 1):
            start = road[0][i]
            end = road[0][i + 1]

            # Calculate the distance between the two points
            dist_m = np.sqrt((start[0] - end[0])**2 + (start[1] - end[1])**2)

            # If the distance is greater than max_spacing_m, interpolate points
            if dist_m > max_spacing_m:
                # Calculate the number of points to interpolate
                num_points = int(np.ceil(dist_m / max_spacing_m))

                # Interpolate the x and y coordinates
                x = np.linspace(start[0], end[0], num_points)
                y = np.linspace(start[1], end[1], num_points)

                # Add the interpolated points to the new road
                interpolated_road.extend(list(zip(x, y)))
            else:
                # If no interpolation is needed, just add the start point
                interpolated_road.append(start)

        # Add the last point of the road
        interpolated_road.append(road[0][-1])

        # Add the new road to the list of roads
        interpolated_roads.append((interpolated_road, road[1]))

    return interpolated_roads


def parse_elevation_data(top_path: str, data_res: float, cell_size: float) -> dict:
    """Read elevation data file, rotate and buffer, and interpolate so it can be used for smaller
    cell sizes.

    :param top_path: path to the elevation data file
    :type top_path: str
    :param data_res: resolution of the raw elevation data
    :type data_res: float
    :param cell_size: size the elevation data will interpolate to (new resolution)
    :type cell_size: float
    :return: dictionary containing all relevant topography data
    :rtype: dict
    """
    with rasterio.open(top_path) as elev_data:
        # read data
        array = elev_data.read(1)

    # Take out no data values
    no_data_value = elev_data.nodatavals[0]

    topography_map = rotate_and_buffer_data(array, no_data_value)

    width_m = topography_map.shape[1] * data_res
    height_m = topography_map.shape[0] * data_res

    x = np.arange(0, topography_map.shape[1])
    y = np.arange(0, topography_map.shape[0])
    X, Y = np.meshgrid(x, y)
    plt.contour(X, Y, np.flipud(topography_map), colors='k')

    # upscale grid
    scale_factor = int(np.floor(data_res/cell_size))

    x = np.arange(topography_map.shape[1])
    y = np.arange(topography_map.shape[0])

    # construct the interpolator object
    interpolator = RectBivariateSpline(y, x, topography_map)

    xnew = np.linspace(0, topography_map.shape[1]-1, topography_map.shape[1]*scale_factor)
    ynew = np.linspace(0, topography_map.shape[0]-1, topography_map.shape[0]*scale_factor)

    # apply the interpolator
    topography_map_new = interpolator(ynew, xnew)
    topography_map_new = topography_map_new.reshape((topography_map.shape[0]*scale_factor,
                                                     topography_map.shape[1]*scale_factor))

    rows = topography_map_new.shape[0]
    cols = topography_map_new.shape[1]
    resolution = width_m / cols

    topography_output = {'width_m': width_m,
                        'height_m': height_m,
                        'rows': rows,
                        'cols': cols,
                        'resolution': resolution,
                        'map': topography_map_new,
                        'uniform': False
                        }

    return topography_output, elev_data.bounds

def create_uniform_elev_map(rows: int, cols: int) -> dict:
    """Create an elevation map for the uniform case (all elevation set to 0)

    :param rows: number of rows to populate data for
    :type rows: int
    :param cols: number of columns to populate data for
    :type cols: int
    :return: dictionary containing relevant topography data for uniform case
    :rtype: dict
    """
    topography_map = np.full((rows, cols), 0)

    topography_output = {'width_m': cols,
                    'height_m': rows,
                    'rows': rows,
                    'cols': cols,
                    'resolution': 1,
                    'map': topography_map,
                    'uniform': True
                    }

    return topography_output

def parse_fuel_data(fuel_path: str, import_roads: bool) -> dict:
    """Read fuel data file, rotate and buffer, outputs dictionary with relevant data

    :param fuel_path: path to the raw fuel data file
    :type fuel_path: str
    :param import_roads: boolean to indicate whether roads will be imported, if True function will
                         replace 'Urban' fuel type in raw data with nearby fuels
    :type import_roads: bool
    :raises ValueError: if fuel data contains values that are not one of the 13 Anderson FBFMs
    :return: dictionary with all relevant fuel data
    :rtype: dict
    """

    with rasterio.open(fuel_path) as fuel_data:
        array = fuel_data.read(1)
    no_data_value = fuel_data.nodatavals[0]

    fuel_map = rotate_and_buffer_data(array, no_data_value, order=0)

    if import_roads:
        fuel_map = replace_clusters(fuel_map)

    for i in range(fuel_map.shape[0]):
        for j in range(fuel_map.shape[1]):
            if fuel_map[i, j] not in fc.fbfm_13_keys:
                raise ValueError("One or more of the fuel values not valid for FBFM13")

    width_m = fuel_map.shape[1] * DATA_RESOLUTION
    height_m = fuel_map.shape[0] * DATA_RESOLUTION

    rows = fuel_map.shape[0]
    cols = fuel_map.shape[1]

    # Create a color list in the right order
    colors = [fc.fuel_color_mapping[key] for key in sorted(fc.fuel_color_mapping.keys())]

    # Create a colormap from the list
    cmap = ListedColormap(colors)

    # Create a norm object to map your data points to the colormap
    norm = BoundaryNorm(list(sorted(fc.fuel_color_mapping.keys())) + [100], cmap.N)

    plt.imshow(np.flipud(fuel_map), cmap=cmap, norm=norm)

    resolution = DATA_RESOLUTION

    fuel_output = {
                    'width_m': width_m,
                    'height_m': height_m,
                    'rows': rows,
                    'cols': cols,
                    'resolution': resolution,
                    'map': fuel_map,
                    'uniform': False
                }

    return fuel_output, fuel_data.bounds

def create_uniform_fuel_map(height_m: float, width_m: float, fuel_type: fc.fbfm_13_keys) -> dict:
    """Create a fuel map for the uniform case (all fuel set to same value)

    :param height_m: height in meters of the map that should be generated
    :type height_m: float
    :param width_m: width in meters of the map that should be generated
    :type width_m: float
    :param fuel_type: fuel type that should be applied across the entire map
    :type fuel_type: FuelConstants.fbfm_13_keys
    :return: dictionary containing all the necessary data for the uniform fuel case
    :rtype: dict
    """
    fuel_map = np.full((int(np.floor(height_m/DATA_RESOLUTION)),
                        int(np.floor(width_m/DATA_RESOLUTION))),
                        int(fuel_type))

    # Create a color list in the right order
    colors = [fc.fuel_color_mapping[key] for key in sorted(fc.fuel_color_mapping.keys())]

    # Create a colormap from the list
    cmap = ListedColormap(colors)

    # Create a norm object to map your data points to the colormap
    norm = BoundaryNorm(list(sorted(fc.fuel_color_mapping.keys())) + [100], cmap.N)

    plt.imshow(fuel_map, cmap=cmap, norm=norm)

    fuel_output = {
                    'width_m': width_m,
                    'height_m': height_m,
                    'rows': int(np.floor(height_m/DATA_RESOLUTION)),
                    'cols': int(np.floor(width_m/DATA_RESOLUTION)),
                    'resolution': DATA_RESOLUTION,
                    'map': fuel_map,
                    'uniform': True,
                    'fuel type': fc.fuel_names[fuel_type]
                }

    return fuel_output

def rotate_and_buffer_data(array: np.ndarray, no_data_value: int, order = 3) -> np.ndarray:
    """Rotates and buffers raw .tif fuel and elevation files so they are squares and contain no
    garbage data

    :param array: array of data to be cleaned
    :type array: np.ndarray
    :param no_data_value: value that represents no data within the array
    :type no_data_value: int
    :param order: order of interpolation for ndimage.rotate, order=0 does not interpolate, defaults
                  to 3
    :type order: int, optional
    :return: cleaned array
    :rtype: np.ndarray
    """
    array = np.where(array == no_data_value, -100, array)

    # find top left corner
    for x in range(array.shape[0]):
        if array[0][x] != -100:
            corner_1 = (x, 0)
            break

    # find bottom left corner
    for y in range(array.shape[1]):
        if array[y][0] != -100:
            corner_2 = (0, y)
            break

    # rotate data to align it with x, y axes
    angle = np.arctan(corner_1[0]/corner_2[1]) * (180/np.pi)
    rotated_data = ndimage.rotate(array, angle, order=order)

    # Get the indices of valid elevation values
    valid_indices = np.where((rotated_data > 0))

    # Get rough min and max valid indices
    min_row = np.min(valid_indices[0])
    max_row = np.max(valid_indices[0])
    min_col = np.min(valid_indices[1])
    max_col = np.max(valid_indices[1])

    # Trim data
    new_array = rotated_data[min_row:max_row, min_col:max_col]

    # buffer 5 values on each side to remove any garbage data
    new_array = new_array[5:-5, 5:-5]

    return new_array

def replace_clusters(fuel_map: np.ndarray, invalid_value=91) -> np.ndarray:
    """Function to replace clusters of invalid values with their neighboring values

    :param fuel_map: 2d array containing fuel data
    :type fuel_map: np.ndarray
    :param invalid_value: value to be replaced, defaults to 91 ('Urban')
    :type invalid_value: int, optional
    :return: 2d array containing fuel data with 'invalid_value' replaced
    :rtype: np.ndarray
    """
    # Check if fuel_map contains any invalid_value
    if invalid_value not in fuel_map:
        return fuel_map

    # Generate a mask for the invalid cells
    invalid_mask = fuel_map == invalid_value

    # Label each cluster of invalid cells
    labels, num_labels = ndimage.label(invalid_mask)

    # Create a dilation structuring element (SE)
    selem = ndimage.generate_binary_structure(2,2)  # 2x2 square SE

    # Initialize an output array to store the final fuel_map
    output_map = fuel_map.copy()

    # Process each label (cluster of invalid cells)
    for label in range(1, num_labels + 1):
        # Create a mask for this label only
        label_mask = labels == label

        # Dilation operation: for this label, expand it to its neighbors
        dilated_mask = ndimage.binary_dilation(label_mask, structure=selem)

        # Exclude the original label cells from the dilated mask
        outer_ring_mask = np.logical_and(dilated_mask, np.logical_not(label_mask))

        # Extract the values of the cells in the outer ring
        outer_ring_values = fuel_map[outer_ring_mask]

        # Exclude any invalid values in the outer ring
        outer_ring_values = outer_ring_values[outer_ring_values != invalid_value]

        # Find the most common value
        if len(outer_ring_values) > 0:
            most_common = stats.mode(outer_ring_values, keepdims=True)[0][0]
        else:
            most_common = invalid_value

        # Replace the label cells in the output_map with the most common value
        output_map[label_mask] = most_common

    return output_map

def get_user_data(fig: matplotlib.figure.Figure) -> dict:
    """Function that generates GUI for user to specify initial ignitions and fire-breaks. Returns
    dictionary containing user inputs.

    :param fig: figure object to deploy GUI in
    :type fig: matplotlib.figure.Figure
    :return: dictionary with all user input from GUI
    :rtype: dict
    """
    user_data = {}

    # display map
    drawer = PolygonDrawer(fig)
    plt.show()

    if not drawer.valid:
        print("Incomplete data provided. Not writing data to file, terminating...")
        sys.exit(0)

    polygons = drawer.get_ignitions()
    transformed_polygons = transform_polygons(polygons)
    shapely_polygons = get_shapely_polys(transformed_polygons)

    polygons = [mapping(polygon) for polygon in shapely_polygons]

    user_data["initial_ignition"] = polygons

    lines, fuel_vals = drawer.get_fire_breaks()

    transformed_lines = transform_lines(lines, DATA_RESOLUTION)

    line_strings = [LineString(line) for line in transformed_lines]

    user_data["fire_breaks"] = [{"geometry": mapping(line), "fuel_value": fuel_value} for line, fuel_value in zip(line_strings, fuel_vals)]

    return user_data

def transform_polygons(polygons: list) -> list:
    """Function to transform polygons to the proper scale for the sim map

    :param polygons: list of polygons drawn by user
    :type polygons: list

    :return: list of polygons scaled up appropriately
    :rtype: list
    """
    transformed_polygons = []
    for polygon in polygons:

        transformed_polygon = [(x*DATA_RESOLUTION, y*DATA_RESOLUTION) for x,y in polygon]
        transformed_polygons.append(transformed_polygon)

    return transformed_polygons

def transform_lines(line_segments: list, scale_factor: float) -> list:
    """Function to transform lines to the proper scale for the sim map

    :param line_segments: list of lines drawn by user
    :type line_segments: list
    :param scale_factor: data resolution shown in drawing GUI, used to scale the line up
                         so they reflect the actually distances they were representing
    :type scale_factor: float
    :return: list of lines scaled up appropriately
    :rtype: list
    """
    transformed_lines = []
    for line in line_segments:
        transformed_line = []
        for pt in line:
            transformed_pt = [pt[0] * scale_factor, pt[1] * scale_factor]
            transformed_line.append(transformed_pt)

        transformed_line = remove_consec_duplicates(transformed_line)

        transformed_lines.append(transformed_line)

    return transformed_lines

def remove_consec_duplicates(line: list) -> list:
    """Function to remove consecutive duplicate points in a line

    :param line: list of (x,y) points that make up a line
    :type line: list
    :return: list of (x,y) points that make up a line with any duplicates removed
    :rtype: list
    """
    cleaned_line = [line[0]]  # start with the first point
    for point in line[1:]:  # for each subsequent point
        if point != cleaned_line[-1]:  # if it's not the same as the last point we added
            cleaned_line.append(point)  # add it to the list
    return cleaned_line

def get_shapely_polys(polygons: list) -> list:
    """Convert list of polygons as generated by map drawer to a list of Shapely polygons

    :param polygons: list of polygons each defined as [(x1,y1),(x2,y2)...]
    :type polygons: list
    :return: list of equivalent polygons as shapely.Polygon objects
    :rtype: list
    """
    shapely_polygons = [Polygon(coords) for coords in polygons]
    return shapely_polygons

def main():
    file_selector = MapGenFileSelector()
    file_params = file_selector.run()

    if file_params is None:
        print("User exited before submitting necessary files.")
        sys.exit(0)

    generate_map_from_file(file_params, DATA_RESOLUTION, 1)

if __name__ == "__main__":
    main()



"""Demo script showing how to access the state of a sim at any snapshot in time
"""
import pickle
import msgpack
import itertools
from utilities.fire_util import UtilFuncs
from utilities.fire_util import CellStates

# Paths to init_fire_state.pkl and pkl log life
INIT_STATE = '/path to log folder/init_fire_state.pkl'
LOG_FILE = '/path to log folder/run_#/log.msgpack'

TIME = 1000 # snapshot in time of interest, make sure this is in range of (0-sim_duration) and that
            # it is an integer multiple of time_step

# Load the initial state
with open(INIT_STATE, 'rb') as f:
    sim = pickle.load(f)

# Retrieve properties from sim dictionary
wind = sim['wind_vec']
cell_dict = sim['cell_dict'] 
cell_grid = sim['cell_grid']

cell_size = sim['cell_size']
grid_width = sim['grid_width']
grid_height = sim['grid_height']

# Load the log file
with open(LOG_FILE, 'rb') as f:
    data = msgpack.unpack(f, strict_map_key=False) # Note: you must use strict_map_key=False when unpacking

time_steps = data.keys() # gives a list of all the time steps that occurred during the simulation

for time in time_steps:
    if time > TIME:
        break

    updated_cells = data[time] # get the cells that were updated during this particular time step

    wind._update_wind(time) # update the wind vector at this time step if that is of interest

    for cell in updated_cells:
        c = cell_dict[cell['id']] # get the cell object

        c._set_state(cell['state'])
        if c.state == CellStates.FIRE:
            c._set_fire_type(cell['fire_type'])
        c._set_fuel_content(cell['fuel_content'])

# cell_dict now contains a dictionary of all the cells at time 't', each key is the id of the cell
# cell_grid contains the cells in grid form at time 't'

# Cells can be accessed using get_indices_from_xy() by providing a position in the xy plane
# or directly by row,col
row, col = UtilFuncs.get_indices_from_xy(3970, 22.5, cell_size, grid_width, grid_height)
cell = cell_grid[row, col]

print(f"cell: {cell}")
print(f"cell size: {cell_size}")



"""Wind representation that defines the wind conditions for a simulation and calculates the effect
on fire propagation

.. autoclass:: Wind
    :members:
"""

from typing import Tuple
import numpy as np

from embrs.utilities.fire_util import WindAdjustments
from embrs.fire_simulator.cell import Cell

class Wind:
    """Wind class that specifies the wind conditions for a given simulation.

    :param forecast: list of tuples representing the wind vector at equally spaced instants in
                     time. Each element represents (speed (m/s), direction (deg)) at an instant.
    :type forecast: list
    :param time_step: time step between each point in 'forecast' (in minutes)
    :type time_step: float
    """

    def __init__(self, forecast: list, time_step: float):
        """Constructor method for a wind instance, creates an initial wind vector and saves input
        parameters.
        """
        print("Configuring wind...")
        wind_dir_rad = forecast[0][1] * (np.pi/180)
        wind_mag_m_s = forecast[0][0]
        self._vec = [wind_mag_m_s*np.cos(wind_dir_rad), wind_mag_m_s*np.sin(wind_dir_rad)]
        self._wind_speed = wind_mag_m_s
        self._wind_dir_deg = forecast[0][1]
        self._time_step = time_step
        self._forecast = forecast
        self._curr_index = 0

    def _update_wind(self, curr_time_s: int) -> bool:
        """Updates the wind to reflect the forecasted conditions at the inputted time.

        :param curr_time_s: current sim time (in seconds).
        :type curr_time_s: int
        """

        curr_time_m = curr_time_s / 60
        curr_index = int(np.floor(curr_time_m / self.time_step))

        if curr_index != self._curr_index:
            self._curr_index = curr_index
            self._wind_speed = self.forecast[self.curr_index][0]
            self._wind_dir_deg = self.forecast[self.curr_index][1]
            wind_dir_rad = self.wind_dir_deg * (np.pi/180)
            self._vec = [self.wind_speed*np.cos(wind_dir_rad),self.wind_speed*np.sin(wind_dir_rad)]

            return True
        return False

    def _calc_wind_effect(self, curr_cell: Cell, disp: Tuple[float, float]) -> Tuple[float, float]:
        """Calculate the effect the wind has on the spread probability from 'cell' to one of its
        neighbors at a displacement 'disp' away.

        :param curr_cell: :class:`~fire_simulator.fire.Cell` instance that is currently on fire
        :type curr_cell: :class:`~fire_simulator.fire.Cell`
        :param disp: tuple representing the delta in col, row from 'curr_cell' to one of its
                     neighbors
        :type disp: Tuple[float, float]
        :return: a tuple of form (alpha_w, k_w). 
        
        - 'alpha_w' is used to weight the probability of ignition.
        - `k_w' is used to weight the propagation velocity of the fire.
        :rtype: Tuple[float, float]
        """
        disp_vec = curr_cell.mapping[disp]

        dot_proj = sum(a*b for a, b in zip(disp_vec, self.vec))

        k_w = max(0, np.exp(0.1783*dot_proj) - 0.486)

        if self.wind_speed == 0:
            alpha_w = 1

        else:
            cos_rel_angle = dot_proj / (np.linalg.norm(disp_vec) * np.linalg.norm(self.vec))
            rel_angle = np.arccos(cos_rel_angle)
            rel_angle = np.degrees(rel_angle)
            adj_vel_kmh = max(self.wind_speed * 3.6 - 8, 0)
            alpha_w = self.interpolate_wind_adjustment(adj_vel_kmh, rel_angle)

        return alpha_w, k_w

    def interpolate_parameters(self, wind_speed: float) -> Tuple[float, float, float]:
        """Calculates the parametes for the Lorentzian function based on the input wind speed.
        Interpolates a series of Lorentzian fits to determine the parameters.

        :param wind_speed: wind speed in km/h to find parameters for
        :type wind_speed: float
        :return: Peak amplitude (A), Full width at half maximum (gamma), and constant offset (C)
                 of the Lorentzian
        :rtype: Tuple[float, float, float]
        """

        param_mapping = WindAdjustments.wind_speed_param_mapping

        sorted_speeds = sorted(param_mapping.keys())

        if wind_speed in sorted_speeds:
            return param_mapping[wind_speed]

        # Find two wind speeds closest to the given wind speed for interpolation
        v_lower = max([s for s in sorted_speeds if s <= wind_speed])
        v_upper = min([s for s in sorted_speeds if s >= wind_speed])

        # Extract parameters for the two closest wind speeds
        A_lower, gamma_lower, C_lower = param_mapping[v_lower]
        A_upper, gamma_upper, C_upper = param_mapping[v_upper]

        # Calculate weights for interpolation
        w1 = (v_upper - wind_speed) / (v_upper - v_lower)
        w2 = 1 - w1

        # Interpolate the parameters
        A_interp = w1 * A_lower + w2 * A_upper
        gamma_interp = w1 * gamma_lower + w2 * gamma_upper
        C_interp = w1 * C_lower + w2 * C_upper

        return A_interp, gamma_interp, C_interp

    def lorentzian(self, x: float, A: float, gamma: float, C: float) -> float:
        """Calculate the output of a Lorentzian function with the given input parameters.

            :param x: Point to calculate the output of the Lorentzian for, representing the
                      relative angle in degrees.
            :type x: float
            :param A: Peak amplitude of the Lorentzian peak.
            :type A: float
            :param gamma: Scale parameter related to the full width at half maximum (FWHM) of the
                          peak.
            :type gamma: float
            :param C: Constant baseline offset of the Lorentzian peak.
            :type C: float
            :return: The value of the Lorentzian function at point x.
            :rtype: float
        """
        return A / (1 + (x/gamma)** 2) + C

    def interpolate_wind_adjustment(self, wind_speed_kmh: float, direction: float) -> float:
        """Calculates the alpha_w parameter which models the probability adjustment due to the
        wind. Interpolates a series of lorentzian functions that define the adjustment factor.

        :param wind_speed_kmh: wind speed in km/h
        :type wind_speed_kmh: float
        :param direction: direction of wind in degrees relative to the propagation direction being
                          considered
        :type direction: float
        :return: wind adjustment factor, alpha_w
        :rtype: float
        """

        # Interpolate the parameters for the given wind speed
        A, gamma, C = self.interpolate_parameters(wind_speed_kmh)

        # Calculate the interpolated value using the Lorentzian function
        return self.lorentzian(direction, A, gamma, C)

    @property
    def vec(self) -> list:
        """Current wind conditions as a 2D vector [x velocity (m/s), y velocity (m/s)].
        """
        return self._vec

    @property
    def wind_speed(self) -> float:
        """Current speed of the wind (m/s).
        """
        return self._wind_speed

    @property
    def wind_dir_deg(self) -> float:
        """Current direction of the wind (deg).
        """
        return self._wind_dir_deg

    @property
    def time_step(self) -> float:
        """Time step between each point in the wind's 'forecast' (in minutes).
        """
        return self._time_step

    @property
    def forecast(self) -> list:
        """List of tuples representing the wind vector at equally spaced instants in
        time. 
        
        Each element represents (speed (m/s), direction (deg)) at an instant. 
        """
        return self._forecast

    @property
    def curr_index(self) -> int:
        """Current index the wind is on within its forecast.
        """
        return self._curr_index



"""Module containing base class for agents

.. autoclass:: AgentBase
    :members:

"""

class AgentBase:
    """Base class for agents in user code. 
    
    Agent objects must be an instance of this class in order to be registered with the sim and
    displayed in visualizations.

    :param id: unique identifier of the agent
    :type id: any
    :param x_m: x position in meters of the agent within the sim.
    :type x_m: float
    :param y_m: y position in meters of the agent within the sim.
    :type y_m: float
    :param label: label added to the agent when it is displayed, if `None` no label will be
                  added, defaults to `None`.
    :type label: str, optional
    :param marker: marker used to represent the agent when displayed, any
                   :py:attr:`matplotlib.markers` can be used, defaults to '*'.
    :type marker: str, optional
    :param color: color used to represent the agent when displayed, any 
                  :py:attr:`matplotlib.colors` can be used, defaults to 'magenta'.
    :type color: str, optional
    """

    def __init__(self, id: any, x: float, y: float, label:str=None, marker:str='*',
                color:str='magenta'):
        """Constructor method that sets the basic parameters for an agent
        """
        self.id = id
        self.x = x
        self.y = y
        self.label = label
        self.marker = marker
        self.color = color

    def to_log_format(self) -> dict:
        """Returns the agent in a format that the logger can store

        :return: dictionary with the following fields:

            - "id": unique identifier of the agent
            - "x_m": x position in meters of the agent within the sim.
            - "y_m": y position in meters of the agent within the sim.
            - "marker": marker used to represent the agent when displayed.
            - "color": color used to represent the agent when displayed.
        :rtype: dict
        """
        data = {
            "id": self.id,
            "label": self.label,
            "x": self.x,
            "y": self.y,
            "marker": self.marker,
            "color": self.color
        }

        return data



"""Representation of the discrete cells that make up the fire sim.

.. autoclass:: Cell
    :members:

"""
import numpy as np
from shapely.geometry import Polygon

from embrs.utilities.fire_util import ControlledBurnParams
from embrs.utilities.fire_util import CellStates, FireTypes
from embrs.utilities.fire_util import HexGridMath as hex
from embrs.fire_simulator.fuel import Fuel

class Cell:
    """Class representation of the discrete cells that make up the fire sim. This holds the state
    of each cell as well as the relevant fire spread parameters. Cells are regular hexagons in the
    "point-up" orientation.

    :param id: unique id of the cell within the sim
    :type id: int
    :param col: column index of the cell within the :py:attr:`~fire_simulator.fire.FireSim.cell_grid`
    :type col: int
    :param row: row index of the cell within the :py:attr:`~fire_simulator.fire.FireSim.cell_grid`
    :type row: int
    :param cell_size: size of the cell, edge length of the hexagons, units: meters
    :type cell_size: float
    :param z: elevation of the cell measured in meters, defaults to 0.0
    :type z: float, optional
    :param fuel_type: type of fuel present at the cell, can be any of the 13 Anderson FBFMs,
                      defaults to Fuel(1)
    :type fuel_type: :class:`~fire_simulator.fuel.Fuel`, optional
    """

    def __init__(self, id: int, col: int, row: int, cell_size: float, z = 0.0, fuel_type=Fuel(1)):
        """Constructor method to initialize a cell instance. Calculates x,y position in Cartesian
        plane.
        """
        self.id = id

        # Set cell indices
        self._col = col
        self._row = row

        # z is the elevation of cell in m
        self._z = z

        self._cell_size = cell_size # defined as the edge length of hexagon
        self._cell_area = self.calc_cell_area()

        # x_pos, y_pos are the global position of cell in m
        if row % 2 == 0:
            self._x_pos = col * cell_size * np.sqrt(3)
        else:
            self._x_pos = (col + 0.5) * cell_size * np.sqrt(3)

        self._y_pos = row * cell_size * 1.5

        self._fuel_type = fuel_type
        self._fuel_content = fuel_type.init_fuel
        self.fuel_at_ignition = None
        self.W = None
        self.ignition_clock = 0

        # Variables to track how long a given cell will take to burnout
        self._t_d = None

        # Variable that tracks which type of fire a burning cell is
        self._fire_type = -1

        self._state = fuel_type.init_state
        self.cont_state = 0 # continuous state, only used for FirePrediction model

        self.has_neighbor = True
        self._neighbors = []
        self._burnable_neighbors = []

        # dead fuel moisture at this cell, value based on Anderson fuel model paper
        self._dead_m = 0.08

        self.polygon = self.to_polygon()
        self.changed = False
        self.mapping = None

        if self._row % 2 == 0:
            self.mapping = hex.even_neighborhood_mapping

        else:
            self.mapping = hex.odd_neighborhood_mapping

    def _set_elev(self, z: float):
        """Set the elevation of a cell

        :param z: elevation in meters
        :type z: float
        """

        self._z = z

    def _set_fuel_content(self, fuel_content: float):
        """Set the fuel content remaining in a cell.

        :param fuel_content: amount of fuel remaining in a cell, ranging from 0 to 1
        :type fuel_content: float
        """
        self._fuel_content = fuel_content
        self.changed = True

    def _set_fuel_type(self, fuel_type: Fuel):
        """Set the fuel type of a cell.

        :param fuel_type: Fuel type, one of the 13 Anderson FBFMs
        :type fuel_type: :class:`~fire_simulator.fuel.Fuel`
        """
        self._fuel_type = fuel_type
        self._fuel_content = fuel_type.init_fuel
        self._state = fuel_type.init_state
        self.changed = True

    def _set_vprop(self, v_prop: float):
        """Calculates the time a fire will take to propagate across a cell from the propagation
        velocity

        :param v_prop: propagation velocity of the fire in m/s
        :type v_prop: float
        """

        # Calculate time it takes for fire to propagate across cell
        self._t_d = (2 * self.cell_size) / v_prop # distance across corners used as the distance

    def _set_state(self, state: CellStates):
        """Set the state of the cell to :py:attr:`CellStates.FUEL`, :py:attr:`CellStates.FIRE`,
        or :py:attr:`CellStates.BURNT`

        :param state: state of the cell
        :type state: :class:`~utilities.fire_util.CellStates`
        """
        self._state = state
        self.changed = True

        if state == CellStates.FIRE:
            if self.fuel_type.fuel_type <= 13: # Make sure cell is combustible type
                prop_speed = self.fuel_type.nominal_vel/60
                self.W = self.fuel_type.consumption_factor

                if self.fire_type == FireTypes.PRESCRIBED:
                    prop_speed *= ControlledBurnParams.nominal_vel_adj
                    self.W *= ControlledBurnParams.consumption_factor_adj

                self._set_vprop(prop_speed)

                if self.fuel_content == 1:        
                    self.ignition_clock = 0 - self._t_d # start ignition clock when fire has propagated across entire cell
                    self.fuel_at_ignition = self.fuel_content
                else:
                    adjusted_ignition_clock = -self.fuel_type.consumption_factor * np.log(self.fuel_content)
                    self.ignition_clock = 0 - adjusted_ignition_clock
                    self.fuel_at_ignition = self.fuel_content

    def _set_fire_type(self, fire_type: FireTypes):
        """Set the type of fire at a cell

        :param fire_type: Fire type to set, either :py:attr:`FireTypes.WILD` or :py:attr:`FireTypes.PRESCRIBED`
        :type fire_type: :class:`~utilities.fire_util.FireTypes`
        """
        self._fire_type = fire_type
        self.changed = True

    def _set_dead_m(self, dead_m: float):
        """Set the dead fuel moisture at a cell.
        
        Higher fuel moisture will slow the propagation of fire

        :param dead_m: dead fuel moisture as a fraction (0 to 1)
        :type dead_m: float
        """
        self._dead_m = dead_m
        self.changed = True

    def __str__(self):
        """To string method for cells.

        :return: formatted string representing the cell
        :rtype: str
        """
        return (f"(id: {self.id}, {self.x_pos}, {self.y_pos}, {self.z}, "
                f"type: {self.fuel_type.name}, "
                f"fuel_content: {self.fuel_content}, "
                f"state: {self.state}")

    def calc_cell_area(self):
        """Calculate the area of a cell in meters squared

        :return: area of a cell given in meters squared
        :rtype: float
        """
        area_m2 = (3 * np.sqrt(3) * self.cell_size ** 2) / 2
        return area_m2

    def to_polygon(self):
        """Generate a Shapely polygon representation of the hexagonal cell.

        :return: A Shapely polygon representing the hexagonal cell
        :rtype: shapely.geometry.Polygon
        """
        l = self.cell_size
        x, y = self.x_pos, self.y_pos

        # Define the vertices for the hexagon in point-up orientation
        hex_coords = [
            (x, y + l),
            (x + (np.sqrt(3) / 2) * l, y + l / 2),
            (x + (np.sqrt(3) / 2) * l, y - l / 2),
            (x, y - l),
            (x - (np.sqrt(3) / 2) * l, y - l / 2),
            (x - (np.sqrt(3) / 2) * l, y + l / 2),
            (x, y + l)  # Close the polygon
        ]

        return Polygon(hex_coords)

    def to_log_format(self):
        """Returns a dictionary of cell data at the current instant.
        Contains all data necessary to capture changes to the state of a cell over time. 
        Used to create logs of the sim for visualization or post-processing use.

        :return: **Dictionary with the following fields:**

                - "id": (int) representing the id of the cell object
                - "state": (:class:`~utilities.fire_util.CellStates`) representing the current state of the cell
                - "fuel_content": (float) representing the amount of fuel remaining in the cell
        
                **If the cell's state is on fire the dictionary will also include:**
                    
                - "fire_type": (:class:`~utilities.fire_util.FireTypes`) representing type of fire in the cell

        :rtype: dict
        """
        cell_data = {
            "id": self.id,
            "state": self._state,
            "fuel_content": self._fuel_content,
            "dead_m": self._dead_m
        }

        if self.state == CellStates.FIRE:
            cell_data['fire_type'] = self._fire_type

        return cell_data

    # ------ Compare operators overloads ------ #
    def __lt__(self, other):
        return self.id < other.id

    def __gt__(self, other):
        return self.id > other.id

    @property
    def col(self) -> int:
        """Column index of the cell within the :py:attr:`~fire_simulator.fire.FireSim.cell_grid`
        """
        return self._col

    @property
    def row(self) -> int:
        """Row index of the cell within the :py:attr:`~fire_simulator.fire.FireSim.cell_grid`
        """
        return self._row

    @property
    def cell_size(self) -> float:
        """Size of the cell in meters.
        
        Measured as the side length of the hexagon.
        """
        return self._cell_size

    @property
    def cell_area(self) -> float:
        """Area of the cell, measured in meters squared.
        """
        return self._cell_area

    @property
    def x_pos(self) -> float:
        """x position of the cell within the sim, measured in meters.
        
        x values increase left to right in the sim visualization window
        """
        return self._x_pos

    @property
    def y_pos(self) -> float:
        """y position of the cell within the sim, measured in meters.
        
        y values increase bottom to top in the sim visualization window
        """
        return self._y_pos

    @property
    def z(self) -> float:
        """Elevation of the cell measured in meters
        """
        return self._z

    @property
    def fuel_type(self) -> Fuel:
        """Type of fuel present at the cell.
        
        Can be any of the 13 Anderson FBFMs
        """
        return self._fuel_type

    @property
    def fuel_content(self) -> float:
        """Fraction of fuel remaining at the cell, between 0 and 1
        """
        return self._fuel_content

    @property
    def t_d(self) -> float:
        """Estimate of the time (in seconds) it will take the current fire to propagate across the
        cell. 

        `-1` if the cell is not on fire
        """
        return self._t_d

    @property
    def state(self) -> CellStates:
        """Current state of the cell.
        
        Can be :py:attr:`CellStates.FUEL`, :py:attr:`CellStates.BURNT`, or :py:attr:`CellStates.FIRE`.
        """
        return self._state

    @property
    def fire_type(self) -> FireTypes:
        """Current type of fire burning at the cell.
        
        Either :py:attr:`FireTypes.PRESCRIBED` or :py:attr:`FireTypes.WILD`, `-1` if not currently in state :py:attr:`CellStates.FIRE`.
        """
        return self._fire_type

    @property
    def neighbors(self) -> list:
        """List of cells that are adjacent to the cell.
        
        Each list element is in the form (id, (dx, dy))
        
        - "id" is the id of the neighboring cell
        - "(dx, dy)" is the difference between the column and row respectively of the cell and its neighbor
        """
        return self._neighbors

    @property
    def burnable_neighbors(self) -> set:
        """Set of cells adjacent to the cell which are in a burnable state.

        Each element is in the form (id, (dx, dy))
        
        - "id" is the id of the neighboring cell
        - "(dx, dy)" is the difference between the column and row respectively of the cell and its neighbor
        """
        return self._burnable_neighbors

    @property
    def dead_m(self) -> float:
        """Dead fuel moisture of the cell, as a fraction (0 to 1)
        """
        return self._dead_m



"""Script demonstrating how the fuel spread probabilities were generated
"""

import numpy as np

original_probs = {
    "sh": {"sh": 0.375, "gr": 0.35, "fpc": 0.4, "nfp": 0.375},
    "gr": {"sh": 0.475, "gr": 0.475, "fpc": 0.475, "nfp": 0.475},
    "fpc": {"sh": 0.325, "gr": 0.25, "fpc": 0.35, "nfp": 0.35},
    "nfp": {"sh": 0.1, "gr": 0.075, "fpc": 0.275, "nfp": 0.075}
}

# deltas
F1_D = -0.025
F2_D = 0
F3_D = 0.025
F4_D = 0.025
F5_D = -0.0125
F6_D = 0.0125
F7_D = -0.025
F8_D = -0.025
F9_D = 0.025
F10_D = -0.025
F11_D = 0
F12_D = 0
F13_D = +0.025

delta_mapping = {
    "f1": F1_D, "f2": F2_D, "f3": F3_D, "f4": F4_D, "f5": F5_D, "f6": F6_D, "f7": F7_D,
    "f8": F8_D, "f9": F9_D, "f10": F10_D, "f11": F11_D, "f12": F12_D, "f13": F13_D
}

category_map = {
    "f1": "gr", "f2": "gr", "f3": "gr", "f4": "sh", "f5": "sh", "f6": "sh", "f7": "sh",
    "f8": "nfp", "f9": "nfp", "f10": "fpc", "f11": "nfp", "f12": "fpc", "f13": "fpc"
}

nom_spread_prob_table = {}

for i, fuel1 in enumerate(category_map.keys()):
    print(f"{fuel1} probs:")
    fuel1_category = category_map[fuel1]
    nom_spread_prob_table[i+1] = {}

    for j, fuel2 in enumerate(category_map.keys()):
        fuel2_category = category_map[fuel2]
        original_prob = original_probs[fuel1_category][fuel2_category]
        prob = original_prob + delta_mapping[fuel1] + delta_mapping[fuel2]

        print(f"{j+1}: {np.round(prob, 3)}")

        nom_spread_prob_table[i+1][j+1] = np.round(prob,4)

    print("")
    print("")

print(f"final table: {nom_spread_prob_table}")



"""Module responsible for running the visualization tool that allows users to load log files and
visualize them.
"""
import pickle
import sys
import os
import matplotlib as mpl
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
from matplotlib.animation import FuncAnimation
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.collections import PatchCollection
import matplotlib.colors as mcolors
from matplotlib import cm
from shapely.geometry import LineString
import pandas as pd
import numpy as np
import msgpack

from .utilities.file_io import VizFolderSelector, LoaderWindow
from .utilities.fire_util import CellStates, FireTypes
from .utilities.fire_util import FuelConstants as fc, RoadConstants as rc, UtilFuncs as util


class VisualizationTool:
    """Class for visualizing fire simulations from their log files

    :param data: dictionary of data retrieved from visualization tool's file IO interface
    :type data: dict
    """
    def __init__(self, data: dict):
        """Constructor method that takes in dictionary of visualization data and generates the
        initial figure.
        """
        if data is None:
            sys.exit(0)

        self.loader_window = LoaderWindow('Loading...', max_value=8)
        self.loader_window.set_text("Loading sim data...")

        log_file = data['file']
        self.update_freq_s = data['freq']
        self.scale_km = data['scale_km']
        self.show_legend = data['legend']
        init_location = data['init_location']
        self.has_agents = data['has_agents']

        if self.has_agents:
            agent_file = data['agent_file']

        filename = os.path.basename(log_file)
        run_folder = os.path.dirname(log_file)
        log_folder = os.path.dirname(run_folder)

        # Get initial state of the system
        if init_location:
            init_path = f"{log_folder}/init_fire_state.pkl"
        else:
            init_path = f"{os.path.dirname(log_folder)}/init_fire_state.pkl"

        with open(init_path, 'rb') as f:
            self.sim = pickle.load(f)

        # Get wind object
        self.wind = self.sim['wind_vec']

        grid_width = self.sim['grid_width']
        grid_height = self.sim['grid_height']
        cell_size = self.sim['cell_size']

        width_m = grid_width * np.sqrt(3) * cell_size
        height_m = grid_height * 1.5 * cell_size

        # Get all sim data
        with open(run_folder + '/' + filename, 'rb') as f:
            self.data = msgpack.unpack(f, strict_map_key = False)

        self.loader_window.set_text("Clustering data based on display frequency...")
        # Cluster the data based on update frequency
        self.cluster_data()

        # Get agent data if present
        if self.has_agents:
            with open(agent_file, 'rb') as f:
                self.agent_data = pickle.load(f)

        self.loader_window.increment_progress()
        self.loader_window.set_text("Initializing figure...")

        # Set axis parameters
        h_fig = plt.figure(figsize=(10, 10))
        h_ax = h_fig.add_axes([0.05, 0.05, 0.9, 0.9])
        h_ax.set_aspect('equal')
        h_ax.axis([0, self.sim['cell_size']*self.sim['grid_width']*np.sqrt(3) -
                 (self.sim['cell_size']*np.sqrt(3)/2), 0,
                  self.sim['cell_size']*1.5*self.sim['grid_height'] - (self.sim['cell_size']*1.5)])

        plt.tick_params(left = False,
                        right = False,
                        bottom = False,
                        labelleft = False,
                        labelbottom = False)

        self.loader_window.increment_progress()
        self.loader_window.set_text("Generating contour map...")

        # Generate contour map
        x = np.arange(0, self.sim['grid_width']+1)
        y = np.arange(0, self.sim['grid_height']+1)
        X, Y = np.meshgrid(x, y)

        cont = h_ax.contour(X*self.sim['cell_size']*np.sqrt(3),Y*self.sim['cell_size']*1.5,
                            self.sim['coarse_topography'], colors='k', zorder=2)
        h_ax.clabel(cont, inline=True, fontsize=10, zorder=3)

        self.loader_window.increment_progress()
        self.loader_window.set_text("Adding patches for each cell...")

        # Add low and high polygons to prevent weird color mapping
        low_poly = mpatches.RegularPolygon((-10, -10), numVertices=6,
                                           radius=1/np.sqrt(3),orientation=0)

        high_poly = mpatches.RegularPolygon((-10, -10), numVertices=6,
                                            radius=1/np.sqrt(3),orientation=0)

        fire_patches = [low_poly, high_poly]
        prescribe_patches = [low_poly, high_poly]
        tree_patches = [low_poly, high_poly]
        burnt_patches = []
        fire_breaks = [low_poly, high_poly]
        alpha_arr = [0, 1]
        break_fuel_arr = [0, 1]

        added_colors = set()
        legend_elements = []

        # Draw each cell based on initial states
        for curr_cell in self.sim['cell_dict'].values():
            polygon = mpatches.RegularPolygon((curr_cell.x_pos, curr_cell.y_pos),
                                               numVertices=6, radius=self.sim['cell_size'],
                                               orientation=0)

            if curr_cell.state == CellStates.FUEL:
                color = fc.fuel_color_mapping[curr_cell.fuel_type.fuel_type]
                if color not in added_colors:
                    added_colors.add(color)
                    legend_elements.append(mpatches.Patch(color = color,
                                                          label = curr_cell.fuel_type.name))

                if curr_cell.fuel_content < 1 and curr_cell.fuel_type.fuel_type < 90:
                    fire_breaks.append(polygon)
                    break_fuel_arr.append(curr_cell.fuel_content)

                else:
                    polygon.set(color = color)
                    tree_patches.append(polygon)

            elif curr_cell.state == CellStates.FIRE:
                if curr_cell.fire_type == FireTypes.WILD:
                    fire_patches.append(polygon)
                    alpha_arr.append(curr_cell.fuel_content)
                else:
                    prescribe_patches.append(polygon)

            else:
                burnt_patches.append(polygon)

        # Add all patches to collections
        if tree_patches:
            tree_patches = np.array(tree_patches)
            tree_coll =  PatchCollection(tree_patches, match_original=True)
            h_ax.add_collection(tree_coll)
        if fire_patches:
            fire_patches = np.array(fire_patches)
            fire_coll = PatchCollection(fire_patches, edgecolor='none', facecolor='#F97306')
            fire_coll.set(array=np.array(alpha_arr), cmap=mpl.colormaps["gist_heat"])
            h_ax.add_collection(fire_coll)
        if prescribe_patches:
            prescribe_patches = np.array(prescribe_patches)
            prescribe_coll = PatchCollection(prescribe_patches, edgecolor='none', facecolor='pink')
            h_ax.add_collection(prescribe_coll)
        if burnt_patches:
            burnt_patches = np.array(burnt_patches)
            burnt_coll = PatchCollection(burnt_patches, edgecolor='none', facecolor='k')
            h_ax.add_collection(burnt_coll)

        if len(fire_breaks) > 0:
            breaks_coll = PatchCollection(fire_breaks, edgecolor='none')
            breaks_coll.set(array= break_fuel_arr, cmap=mpl.colormaps["gist_gray"])
            h_ax.add_collection(breaks_coll)

        self.loader_window.increment_progress()
        self.loader_window.set_text("Adding info artists...")

        # Generate time display area
        self.time_box = mpatches.Rectangle((0,self.sim['grid_height']*1.5*self.sim['cell_size']
                                            -(150/6000)*height_m), (1000/6000)*width_m,
                                            (150/6000)*height_m, facecolor='white',
                                            edgecolor='black', linewidth=1, zorder=3, alpha = 0.75)
        h_ax.add_patch(self.time_box)

        sim_time_s = 0
        time_str = util.get_time_str(sim_time_s)

        rx, ry = self.time_box.get_xy()
        cx = rx + self.time_box.get_width()/2
        cy = ry + self.time_box.get_height()/2

        self.simtext = h_ax.text(2*cx - 20, cy, time_str, ha='right', va='center')
        timeheader = h_ax.text(20, cy, 'Time: ', ha='left',va ='center')
        h_ax.add_artist(timeheader)

        # Generate wind display area
        self.wind_box = mpatches.Rectangle((0, self.sim['grid_height']*1.5*self.sim['cell_size']
                                            -(650/6000)*height_m), (500/6000)*width_m,
                                            (500/6000)*height_m, facecolor='white',
                                            edgecolor = 'black', linewidth=1, zorder = 3,
                                            alpha = 0.75)

        h_ax.add_patch(self.wind_box)

        wx, wy = self.wind_box.get_xy()
        cx = wx + self.wind_box.get_width()/2

        windheader = h_ax.text(cx, wy + 0.85 * self.wind_box.get_height(), 'Wind:', ha = 'center',
                               va = 'center')

        h_ax.add_artist(windheader)

        # Generate scale area
        scale_box = mpatches.Rectangle((0, 10), self.scale_km*1000 + 100, (200/6000)*height_m,
                                       facecolor='white', edgecolor='k', linewidth= 1, alpha=0.75,
                                       zorder= 3)
        h_ax.add_patch(scale_box)

        num_cells_scale = self.scale_km * 1000

        if self.scale_km < 1:
            scale_size = str(self.scale_km * 1000) + "m"
        else:
            scale_size = str(self.scale_km) + "km"

        scalebar = AnchoredSizeBar(h_ax.transData, num_cells_scale, scale_size, 'lower left',
                                   color ='k', pad = 0.1, frameon=False)

        h_ax.add_artist(scalebar)

        self.loader_window.increment_progress()

        # Plot roads if they exist
        if 'roads' in self.sim.keys():
            self.loader_window.set_text("Plotting roads...")
            roads = self.sim['roads']

            for road in roads:
                x_trimmed = []
                y_trimmed = []

                x, y = zip(*road[0])
                for i in range(len(x)):
                    if 0 <x[i]/30 <self.sim['grid_width'] and 0 < y[i]/30 <self.sim['grid_height']:
                        x_trimmed.append(x[i])
                        y_trimmed.append(y[i])

                x = tuple(xi for xi in x_trimmed)
                y = tuple(yi for yi in y_trimmed)

                road_color = rc.road_color_mapping[road[1]]

                h_ax.plot(x, y, color= road_color)

                if road_color not in added_colors:
                    added_colors.add(road_color)
                    legend_elements.append(mpatches.Patch(color=road_color,
                                                          label = f"Road - {road[1]}"))


        fire_breaks = self.sim['fire_breaks']

        # Create a colormap for grey shades
        cmap = mpl.colormaps["Greys_r"]

        for fire_break in fire_breaks:
            line = fire_break['geometry']
            fuel_val = fire_break['fuel_value']
            if isinstance(line, LineString):
                # Normalize the fuel_val between 0 and 1
                normalized_fuel_val = fuel_val / 100.0
                color = cmap(normalized_fuel_val)
                x, y = line.xy
                h_ax.plot(x, y, color=color)

        self.loader_window.increment_progress()

        # Add legend area
        if self.show_legend:
            h_ax.legend(handles=legend_elements, loc='upper right', borderaxespad=0)

        # Create class objects
        self.arrow_obj = None
        self.windtext = None
        self.anim = None
        self.static_artists = []
        self.h_ax = h_ax
        self.fig = h_fig

        self.loader_window.increment_progress()
        self.loader_window.set_text("Rendering...")

        # Pause to allow rendering to complete
        plt.pause(1)

        self.loader_window.close()

    def update_viz(self, frame: int) -> list:
        """Function that updates the visualization based on all the changes to the state since the
        last frame

        :param frame: current frame the visualization is on
        :type frame: int
        :return: list of artists drawn
        :rtype: list
        """
        # Create an array to hold elements to be changed
        artists = []

        # Reset static artists if viz has restarted
        if frame == 0:
            self.static_artists = []

        # Get current time in seconds
        curr_time = frame * self.update_freq_s

        # Get updated cells for this iteration
        updated_cells = self.data[frame]

        # Update wind vector
        self.wind._update_wind(curr_time)

        # Add low and high polygons to prevent weird color mapping
        low_poly = mpatches.RegularPolygon((-10, -10), numVertices=6, radius=1/np.sqrt(3),
                                            orientation=0)

        high_poly = mpatches.RegularPolygon((-10, -10), numVertices=6, radius=1/np.sqrt(3),
                                            orientation=0)

        fire_patches = [low_poly, high_poly]
        tree_patches = [low_poly, high_poly]
        burnt_patches = []
        prescribe_patches = [low_poly, high_poly]

        alpha_arr = [0,1]

        soak_xs = []
        soak_ys = []
        c_vals = []

        # Get initial state to apply static properties of cells
        cell_dict = self.sim['cell_dict']

        # Draw all the cells that need to be updated
        for cell in updated_cells:
            c = cell_dict[cell['id']]

            polygon = mpatches.RegularPolygon((c.x_pos, c.y_pos), numVertices=6,
                                              radius=self.sim['cell_size'], orientation=0)

            if cell['state'] == CellStates.FUEL:
                color=np.array(list(mcolors.to_rgba(fc.fuel_color_mapping[c.fuel_type.fuel_type])))
                # Scale color based on cell's fuel content
                color = color *  cell['fuel_content']
                polygon.set_facecolor(color)
                tree_patches.append(polygon)

                if cell['dead_m'] > 0.08: # fuel moisture not nominal
                    soak_xs.append(c.x_pos)
                    soak_ys.append(c.y_pos)
                    c_val = cell['dead_m']/fc.dead_fuel_moisture_ext_table[c.fuel_type.fuel_type]
                    c_val = np.min([1, c_val])
                    c_vals.append(c_val)

            elif cell['state'] == CellStates.FIRE:
                if cell['fire_type'] == FireTypes.WILD:
                    fire_patches.append(polygon)
                    alpha_arr.append(cell['fuel_content'])
                else:
                    prescribe_patches.append(polygon)

            else:
                burnt_patches.append(polygon)

        color_map = cm.get_cmap('Blues')
        norm = mcolors.Normalize(vmin=0, vmax=1)
        moisture_viz = self.h_ax.scatter(soak_xs, soak_ys, c=c_vals, cmap=color_map, marker='2',
                                         norm=norm)

        self.static_artists.append(moisture_viz)

        # Add patches to collections
        if tree_patches:
            tree_patches = np.array(tree_patches)
            tree_coll =  PatchCollection(tree_patches, match_original=True)
            self.h_ax.add_collection(tree_coll)
            artists.append(tree_coll)
        if fire_patches:
            fire_patches = np.array(fire_patches)
            fire_coll = PatchCollection(fire_patches, edgecolor='none', facecolor='#F97306')
            fire_coll.set(array=np.array(alpha_arr), cmap=mpl.colormaps["gist_heat"])
            self.h_ax.add_collection(fire_coll)
            artists.append(fire_coll)
        if prescribe_patches:
            prescribe_patches = np.array(prescribe_patches)
            prescribe_coll = PatchCollection(prescribe_patches, edgecolor='none', facecolor='pink')
            self.h_ax.add_collection(prescribe_coll)
            artists.append(prescribe_coll)
        if burnt_patches:
            burnt_patches = np.array(burnt_patches)
            burnt_coll = PatchCollection(burnt_patches, edgecolor='none', facecolor='k')
            artists.append(burnt_coll)
            self.h_ax.add_collection(burnt_coll)
            self.static_artists.append(burnt_coll)

        # Update wind text
        wx, wy = self.wind_box.get_xy()
        cx = wx + self.wind_box.get_width()/2
        cy = wy + self.wind_box.get_height()/2

        self.windtext = self.h_ax.text(cx, wy + 0.1 * self.wind_box.get_height(),
                                       str(np.round(self.wind.wind_speed,2)) + " m/s",
                                       ha = 'center', va = 'center')

        artists.append(self.windtext)

        # Update wind vector
        if self.wind.wind_speed != 0:
            wind_dir_vec = self.wind.vec/np.linalg.norm(self.wind.vec)
            dx = wind_dir_vec[0]
            dy = wind_dir_vec[1]
            arrow_len = self.wind_box.get_height()/3

            self.arrow_obj = self.h_ax.arrow(cx-(arrow_len*dx) , cy-(arrow_len*dy), dx*arrow_len,
                                             dy*arrow_len, width=10, head_width = 50, color = 'r',
                                             zorder= 3)

        else:
            self.arrow_obj = self.h_ax.text(cx, cy, 'X', fontsize = 20, color = 'r', ha='center',
                                            va = 'center')

        artists.append(self.arrow_obj)

        # Update time string
        time_str = util.get_time_str(curr_time)

        rx, ry = self.time_box.get_xy()
        cx = rx + self.time_box.get_width()/2
        cy = ry + self.time_box.get_height()/2

        # Plot agents for this iteration
        if self.has_agents:
            if curr_time in self.agent_data:
                agents = self.agent_data[curr_time]

                for a in agents:
                    agent_disp = self.h_ax.scatter(a["x"], a["y"], marker=a["marker"],
                                                   color=a["color"])
                    if a["label"] is not None:
                        agent_label = self.h_ax.annotate(str(a["label"]), (a["x"], a["y"]))
                        artists.append(agent_label)
                    artists.append(agent_disp)

        self.simtext.set_visible(False)
        self.simtext = self.h_ax.text(2*cx - 20 , cy, time_str, ha='right', va='center')
        artists.append(self.simtext)

        return artists + self.static_artists

    def run_animation(self):
        """Runs animation based on the time steps of the simulation
        """
        time_steps = self.data.keys()
        self.anim = FuncAnimation(self.fig, self.update_viz, frames=time_steps, interval=1,
                                  blit=True)

        plt.show()

    def cluster_data(self):
        """Function that clusters data into groups of time steps based on the user's selection
        of sim time per frame to be displayed
        """
        # Convert the dictionary into a DataFrame
        df = pd.concat({k: pd.DataFrame(v) for k, v in self.data.items()}, names=['timestamp'])

        # Reset the index to get timestamp as a column
        df.reset_index(level=0, inplace=True)

        # Create a new column 'bins' that will represent the new intervals
        df['bins'] = df['timestamp'] // self.update_freq_s

        # Group by 'id' and 'bins', and keep the last entry in each group
        df = df.groupby(['id', 'bins']).last()

        # Reset index to get back original structure
        df.reset_index(inplace=True)

        # Group by 'bins' and convert each group to a dictionary, then convert to a dictionary
        result = df.groupby('bins').apply(lambda x: x.drop(columns=['bins']).to_dict('records')).to_dict()

        self.data = result

def run(data: dict):
    """Function to run after a user selects the visualization parameters to run. Constructs a 
    VisualizationTool instance with input data and calls the 'run_animation' function

    :param data: _description_
    :type data: dict
    """
    viz = VisualizationTool(data)
    viz.run_animation()

def main():
    folder_selector = VizFolderSelector(run)
    folder_selector.run()

if __name__ == "__main__":
    main()



"""Module for running a prediction on a fire to project its state in a set amount of time.

Runs a simplified version of the fire simulator to provide predictions.

Utilizes a homogenous fuel map, adds noise to the wind forecast, and has the ability to be biased
towards a more or less conservative prediction.

.. autoclass:: FirePredictor
    :members:

"""
from embrs.fire_simulator.fire import FireSim
from embrs.fire_simulator.wind import Wind
from embrs.base_classes.base_fire import BaseFireSim
from embrs.utilities.fire_util import FireTypes, CellStates, UtilFuncs

import numpy as np

class FirePredictor(BaseFireSim):
    """Predictor class responsible for running predictions over a fixed time horizon.

    :param orig_fire: FireSim object to predict the future spread for.
    :type orig_fire: FireSim
    :param time_horizon_hr: Time horizon in hours for which the model should predict over.
    :type time_horizon_hr: float
    :param fuel_type: Fuel type to be used for the predictor's homogenous fuel map. If -1, the
        dominant fuel type in the original map will be used, defaults to -1
    :type fuel_type: int, optional
    :param bias: Bias term that controls how conservative the model is, will tend to over-predict
        if >1, under-predict if <1, defaults to 1
    :type bias: float, optional
    :param time_step_s: Time step in seconds used for the prediction. This will be the temporal 
        granularity of prediction result. If None, the time step of orig_fire will be used.
    :type time_step_s: float
    :param cell_size_m: Cell size in meters used for the prediction. This will be the spatial
        granularity of the prediction result. If None, the cell size of orig_fire will be used.
    :type cell_size_m: float, defaults to None
    """
    def __init__(self, orig_fire: FireSim, time_horizon_hr: float, fuel_type: int = -1, bias: float = 1, time_step_s: float = None, cell_size_m: float = None):
        """Constructor for prediction model.
        """

        if time_step_s is None:
            time_step_s = orig_fire.time_step * 2

        if cell_size_m is None:
            cell_size_m = orig_fire.cell_size * 2

        if not isinstance(fuel_type, int) or fuel_type <= 0 or fuel_type > 13:
            fuel_type = UtilFuncs.get_dominant_fuel_type(orig_fire.base_fuel_map)

            if fuel_type != -1:
                orig_fire.logger.log_message("Invalid fuel type passed to prediction model, defaulted to dominant fuel type in original map")
                print("Invalid fuel type passed to prediction model, defaulted to dominant fuel type in original map")

        self.bias = bias

        fuel_map = np.full(orig_fire.base_fuel_map.shape, fuel_type)
        fuel_res = orig_fire.fuel_res

        topography_map = orig_fire.base_topography
        topography_res = orig_fire.topography_res

        wind_forecast = self.generate_noisy_wind(orig_fire.wind_vec._forecast)
        wind_t_step = orig_fire.wind_vec._time_step

        wind_vec = Wind(wind_forecast, wind_t_step)

        roads = orig_fire.roads
        fire_breaks = orig_fire.fire_breaks

        time_horizon_s = time_horizon_hr * 60 * 60

        # extract initial ignition from curr_fires of orig_fire
        initial_ignition = UtilFuncs.get_cell_polygons(orig_fire.curr_fires)

        sim_size = orig_fire.size
        burnt_cells = UtilFuncs.get_cell_polygons(orig_fire._burnt_cells)

        display_freq_s = orig_fire.display_frequency

        super().__init__(fuel_map, fuel_res, topography_map,
                topography_res, wind_vec, roads, fire_breaks,
                time_step_s, cell_size_m, time_horizon_s, initial_ignition,
                sim_size, burnt_cells, display_freq_s)

        self.start_time_s = orig_fire.curr_time_s
        self._curr_time_s = self.start_time_s

        self._future_fires = {}

    def generate_noisy_wind(self, wind_forecast: list) -> list:
        """Adds noise to the true wind forecast using a auto-regressive model.

        :param wind_forecast: Wind forecast being used by the original fire simulation
        :type wind_forecast: list
        :return: New forecast with noise added
        :rtype: list
        """
        phi = 0.8

        new_forecast = []
        speed_errors = np.zeros(len(wind_forecast))
        dir_errors = np.zeros(len(wind_forecast))

        for i, (speed, dir) in enumerate(wind_forecast):
            if i != 0:
                speed_errors[i] = speed_errors[i - 1] * phi + np.random.normal(0, 1)
                dir_errors[i] = dir_errors[i - 1] * phi + np.random.normal(0, 2.5)

            new_forecast.append((speed + speed_errors[i], dir + dir_errors[i]))

        return new_forecast

    def iterate(self):
        """Iterate the fire prediction one time-step forward.
        """
        self._init_iteration()

        # Update wind
        self.wind_changed = self._wind_vec._update_wind(self._curr_time_s)

        # Iterate over currently burning cells
        for curr_cell in self._curr_fires:

            if len(curr_cell.burnable_neighbors) == 0:
                self._curr_fires_anti_cache.append(curr_cell)
                continue

            # iterate over burnable neighbors
            neighbors_to_remove = []
            for neighbor_id, (dx, dy) in curr_cell.burnable_neighbors:
                neighbor = self._cell_dict[neighbor_id]

                neighbor_burnable = self.check_if_burnable(curr_cell, neighbor)

                if not neighbor_burnable:
                    if neighbor.fire_type != FireTypes.PRESCRIBED:
                        neighbors_to_remove.append((neighbor_id, (dx, dy)))

                else:
                    prob, v_prop = self._calc_prob(curr_cell, neighbor, (dx, dy))

                    prob *= self.bias
                    v_prop *= self.bias

                    if np.random.random() < prob:

                        time_entry = self._future_fires.get(self.curr_time_s)

                        if not time_entry:
                            self._future_fires[self.curr_time_s] = [(neighbor.x_pos, neighbor.y_pos)]

                        else:
                            time_entry.append((neighbor.x_pos, neighbor.y_pos))

                        neighbor._set_vprop(v_prop)
                        self.set_state_at_cell(neighbor, CellStates.FIRE, curr_cell.fire_type)

                        # Add cell to prediction dictionary
                        self._future_fires[self._curr_time_s].append((neighbor.x_pos, neighbor.y_pos))

            curr_cell.burnable_neighbors.difference_update(neighbors_to_remove)

        self._iters += 1

    def _init_iteration(self):
        """Set up an iteration. Resets relevant data structures for keeping track of fires burning
        on the frontier.
        """
        # Update current time
        self._curr_time_s = self.start_time_s + (self.time_step * self.iters)

        self._curr_fires = self._curr_fires | set(self._curr_fires_cache)
        self._curr_fires.difference_update(self._curr_fires_anti_cache)

        self._curr_fires_cache = []
        self._curr_fires_anti_cache = []

        if len(self._curr_fires) == 0 or (self.time_step * self.iters) >= self._sim_duration:
            self._finished = True

    def run_prediction(self) -> dict:
        """Run a prediction

        :return: Dictionary where each key is a time-step and each value is a list of predicted
        ignition locations (x, y) at that time-step. Time-steps start at the time-step the original
            fire when input to the prediction model.
        :rtype: dict
        """

        while not self.finished:
            self.iterate()

        return self._future_fires



import matplotlib.pyplot as plt
import matplotlib.patches as patches

road_color_mapping = {
    'motorway': '#4B0082',
    'trunk': '#800080',
    'primary': '#9400D3',
    'secondary': '#9932CC',
    'tertiary': '#BA55D3',
    'unclassified': '#DA70D6',
    'residential': '#EE82EE',
}

fig = plt.figure(figsize=(12, 3))

for i, (key, color) in enumerate(road_color_mapping.items()):
    rect = patches.Rectangle((i, 0), 1, 1, facecolor=color)
    plt.gca().add_patch(rect)
    plt.text(i+0.5, 0.5, key, color='white', weight='bold',
             horizontalalignment='center', verticalalignment='center')

plt.xlim(0, len(road_color_mapping))
plt.ylim(0, 1)
plt.axis('off')  # Turn off the box around the plot
plt.show()



"""Core fire simulation model.

.. autoclass:: FireSim
    :members:
"""

from tqdm import tqdm
import numpy as np

from embrs.base_classes.base_fire import BaseFireSim
from embrs.utilities.fire_util import CellStates, FireTypes
from embrs.utilities.fire_util import ControlledBurnParams
from embrs.fire_simulator.cell import Cell
from embrs.fire_simulator.wind import Wind

class FireSim(BaseFireSim):
    """Fire simulator class utilizing a probabilistic fire model.
    
    Based on the work of `Trucchia et. al <https://www.mdpi.com/2571-6255/3/3/26>`_. The basis of
    the simulator is a hexagonal grid, each cell in the  grid is a 
    :class:`~fire_simulator.cell.Cell` object that contains all the pertinent fire spread parameters
    for that cell. The simulator takes into account the effects of fuel, slope, and wind on the
    fire propagation.

    :param fuel_map: 2D array that represents the spatial distribution of fuel types in the sim.
                     Each element must be one of the
                     `13 Anderson FBFMs <https://www.fs.usda.gov/rm/pubs_int/int_gtr122.pdf>`_.
                     Note: array is flipped across the horizontal axis after input.
    :type fuel_map: np.ndarray
    :param fuel_res: Resolution of the fuel map in meters. How many meters each row/column
                     represents in space
    :type fuel_res: float
    :param topography_map: 2D array that represents the elevation in meters at each point in space.
                           Note: array is flipped across the
                           horizontal axis after input.
    :type topography_map: np.ndarray
    :param topography_res: Resolution of the topography map in meters. How many meters each
                           row/column represents in space.
    :type topography_res: float
    :param wind_vec: Wind object defining the wind conditions for the simulation.
    :type wind_vec: :class:`~fire_simulator.wind.Wind`
    :param roads: List of points that will be considered roads for the simulation. Format for each
                  element in list: ((x,y), fuel_content). Where (x,y) is a tuple containing the
                  spatial in the sim measured in meters, and fuel_content is the amount of fuel to
                  be modeled at that point (between 0 and 1)
    :type roads: list
    :param fire_breaks: List of dictionaries representing fire-breaks where each dictionary has a
                        "geometry" key with a :py:attr:`shapely.LineString` value and a 
                        "fuel_value" key with a float value which represents the amount of fuel
                        modeled along the :py:attr:`shapely.LineString`.
    :type fire_breaks: list
    :param time_step: Time step of the simulation, defines the amount of time each iteration will
                      model.
    :type time_step: int
    :param cell_size: Size of each cell in the simulation, measured as the distance in meters
                    between two parallel sides of the regular hexagon cells.
    :type cell_size: float
    :param duration_s: Duration of time (in seconds) the simulation should run for, the sim will
                       run for this duration unless the fire is extinguished before the duration
                       has passed.
    :type duration_s: float
    :param initial_ignition: List of shapely polygons that represent the regions of the sim that
                             should start as initially on fire.
    :type initial_ignition: list
    :param size: Size of the simulation backing array (rows, cols)
    :type size: tuple
    :param display_freq_s: The amount of time (in seconds) between updating the real-time
                           visualizer, only used if real-time visualization is selected, defaults
                           to 300
    :type display_freq_s: int, optional
    """
    def __init__(self, fuel_map: np.ndarray, fuel_res: float, topography_map: np.ndarray,
                topography_res: float, wind_vec: Wind, roads: list, fire_breaks: list,
                time_step: int, cell_size: float, duration_s: float, initial_ignition: list,
                size: tuple, display_freq_s = 300):
        """Constructor method to initialize a fire simulation instance. Saves input parameters,
        creates backing array, populates cell_grid and cell_dict with cells, sets initial ignition,
        applies fire-breaks and roads.
        """

        print("Simulation Initializing...")

        # Fuel fraction to be considered BURNT
        self.burnout_thresh = 0.1

        self.logger = None
        self.progress_bar = None

        self._updated_cells = {}
        self._curr_updates = []

        self._partially_burnt = []
        self._soaked = []

        self._agent_list = []
        self._agents_added = False

        self._curr_fires_cache = []
        self._curr_fires_anti_cache = []
        self._curr_fires = set()

        self.w_vals = []
        self.fuels_at_ignition = []
        self.ignition_clocks = []

        super().__init__(fuel_map, fuel_res, topography_map,
                topography_res, wind_vec, roads, fire_breaks,
                time_step, cell_size, duration_s, initial_ignition,
                size, display_freq_s= display_freq_s)

    def iterate(self):
        """Step forward the fire simulation a single time-step
        """
        # Set-up iteration
        if self._init_iteration():
            return

        # Update wind
        self.wind_changed = self._wind_vec._update_wind(self._curr_time_s)

        # Iterate over currently burning cells
        for curr_cell in self._curr_fires:

            # iterate over burnable neighbors
            neighbors_to_remove = []
            for neighbor_id, (dx, dy) in curr_cell._burnable_neighbors:
                neighbor = self._cell_dict[neighbor_id]

                neighbor_burnable = self.check_if_burnable(curr_cell, neighbor)

                if not neighbor_burnable:
                    if neighbor.fire_type != FireTypes.PRESCRIBED:
                        neighbors_to_remove.append((neighbor_id, (dx, dy)))

                else:
                    prob, v_prop = self._calc_prob(curr_cell, neighbor, (dx, dy))

                    if np.random.random() < prob:
                        neighbor._set_vprop(v_prop)
                        self.set_state_at_cell(neighbor, CellStates.FIRE, curr_cell.fire_type)
                        # Add cell to update dictionary
                        self._updated_cells[neighbor.id] = neighbor

                        if neighbor.to_log_format() not in self._curr_updates:
                            self._curr_updates.append(neighbor.to_log_format())


                        if neighbor in self._frontier:
                            self._frontier.remove(neighbor)
                    else:
                        if neighbor not in self._frontier:
                            self._frontier.add(neighbor)

            curr_cell._burnable_neighbors.difference_update(neighbors_to_remove)

            self.capture_cell_changes(curr_cell)

        self.update_fuel_contents()

        self.log_changes()

    def log_changes(self):
        """Log the changes in state from the current iteration
        """
        self._curr_updates.extend(self._partially_burnt)
        self._curr_updates.extend(self._soaked)
        self._soaked = []
        self._iters += 1
        self.logger.add_to_cache(self._curr_updates.copy(), self.curr_time_s)

        if self.agents_added:
            self.logger.add_to_agent_cache(self._get_agent_updates(), self.curr_time_s)

    def update_fuel_contents(self):
        """Update the fuel content of all the burning cells based on the mass-loss algorithm in 
            `(Coen 2005) <http://dx.doi.org/10.1071/WF03043>`_
        """
        curr_fires = self._curr_fires
        burnout_thresh = self.burnout_thresh
        time_step = self.time_step

        ignition_clocks = np.array(self.ignition_clocks)
        fuels_at_ignition = np.array(self.fuels_at_ignition)
        w_vals = np.array(self.w_vals)

        ignition_clocks += time_step
        fuel_contents = fuels_at_ignition * np.minimum(1, np.exp(-ignition_clocks/w_vals))

        relational_dict = self._relational_dict
        updated_cells = self._updated_cells
        curr_updates = self._curr_updates

        for cell, fuel_content, ignition_clock in zip(curr_fires, fuel_contents, ignition_clocks):
            cell._fuel_content = fuel_content
            cell.ignition_clock = ignition_clock

            burnout_thresh = cell.fuel_at_ignition * ControlledBurnParams.burnout_fuel_frac if cell.fire_type == FireTypes.PRESCRIBED else self.burnout_thresh

            if fuel_content < burnout_thresh:
                self.set_state_at_cell(cell, CellStates.BURNT)
                # Add cell to update dictionary
                self._updated_cells[cell.id] = cell

                if cell.to_log_format() not in self._curr_updates:
                    self._curr_updates.append(cell.to_log_format())


                for neighbor_id, _ in cell.neighbors:
                    key = (cell.id, neighbor_id)
                    if key in relational_dict:
                        del relational_dict[key]

            updated_cells[cell.id] = cell
            curr_updates.append(cell.to_log_format())

    def capture_cell_changes(self, curr_cell: Cell):
        """Update data structures with values relating to a burning cell's fuel.
        
        Data structures are later used by :py:attr:`~fire_simulator.fire.update_fuel_contents`

        :param curr_cell: Burning cell to grab values from
        :type curr_cell: Cell
        """
        self.w_vals.append(curr_cell.W)
        self.fuels_at_ignition.append(curr_cell.fuel_at_ignition)
        self.ignition_clocks.append(curr_cell.ignition_clock)

    def _init_iteration(self) -> bool:
        """Set up the next iteration. Reset and update relevant data structures based on last 
        iteration. 

        :return: Boolean value representing if the simulation should be terminated.
        :rtype: bool
        """
        if self._iters == 0:
            self.progress_bar = tqdm(total=self._sim_duration/self.time_step,
                                     desc='Current sim ', position=0, leave=False)

        self._curr_updates.clear()

        # Update current time
        self._curr_time_s = self.time_step * self._iters
        self.progress_bar.update()

        self._curr_fires = self._curr_fires | set(self._curr_fires_cache)
        self._curr_fires.difference_update(self._curr_fires_anti_cache)

        self._curr_fires_cache = []
        self._curr_fires_anti_cache = []

        self.w_vals = []
        self.fuels_at_ignition = []
        self.ignition_clocks = []

        if len(self._curr_fires) == 0 or self._curr_time_s >= self._sim_duration:
            self._finished = True
            self.progress_bar.close()

            if len(self._curr_fires) == 0:
                self.logger.log_message("Fire extinguished! Terminating early.")
            return True

        return False

    def _get_agent_updates(self):
        """Returns a list of dictionaries describing the location of each agent in __agent_list

        :return: List of dictionaries, describing the x,y position of each agent as well as their
                 display preferences
        :rtype: list
        """
        agent_data = []

        for agent in self.agent_list:
            agent_data.append(agent.to_log_format())

        return agent_data

    @property
    def updated_cells(self) -> dict:
        """Dictionary containing cells updated since last time real-time visualization was updated.
        Dict keys are the ids of the :class:`~fire_simulator.cell.Cell` objects.
        """
        return self._updated_cells

    @property
    def curr_updates(self) -> list:
        """List of cells updated during the most recent iteration.
        
        Cells are in their log format as generated by 
        :func:`~fire_simulator.cell.Cell.to_log_format()`
        """
        return self._curr_updates


    @property
    def agent_list(self) -> list:
        """List of :class:`~base_classes.agent_base.AgentBase` objects representing agents
        registered with the sim.
        """

        return self._agent_list

    @property
    def agents_added(self) -> bool:
        """`True` if agents have been registered in sim's agent list, returns `False` otherwise
        """
        return self._agents_added



"""Demo class demonstrating some useful functions of the fire's interface.

To run this example code, start a fire sim and select this file as the "User Module"
"""

from base_classes.control_base import ControlClass
from fire_simulator.fire import FireSim
from utilities.fire_util import CellStates

import numpy as np

class InterfaceDemo(ControlClass):
	def __init__(self, fire:FireSim):

		# Get map size in meters
		width_m, height_m = fire.size
		print(f'Map is {width_m} m x {height_m} m')

		# Save x and y coordinate for later
		self.x = int(np.floor(width_m/4))
		self.y = int(np.floor(height_m/4))

		# Get shape of map's backing array
		rows, cols = fire.shape
		print(f'Map backing array has {rows} rows and {cols} cols')

		# Save the middle row and col for later
		self.row = int(np.floor(rows/2))
		self.col = int(np.floor(cols/2))

		# Get simulation time step
		self.time_step_s = fire.time_step
		print(f'Sim time step is {self.time_step_s}')

		# Variable to keep track of wind updates
		self.last_wind_update = 0

	def process_state(self, fire:FireSim):
		# Check the current sim time and perform cell operations at 30 minutes
		if np.abs((fire.curr_time_m - 30)) == 0:
			print("Performing cell operations")

			# Get the cell containing the saved coordinates
			cell_from_x_y = fire.get_cell_from_xy(self.x, self.y)

			# Get the state of the cell
			cell_state = cell_from_x_y.state

			# Set the fuel content and fuel moisture of the cell if it's unburnt 
			if cell_state == CellStates.FUEL:
				fuel_content = 0.25
				fire.set_fuel_content_at_cell(cell_from_x_y, fuel_content)

				# Get the dead moisture of extinction for the fuel at the cell
				fuel_moisture =  cell_from_x_y.fuel_type.dead_m_ext
				
				# Set the fuel moisture to half the moisture of extinction
				fire.set_fuel_moisture_at_cell(cell_from_x_y, fuel_moisture)

			# Get the cell at the saved indices
			cell_from_row_col = fire.get_cell_from_indices(self.row, self.col)

			# Get the state of the cell
			cell_state = cell_from_row_col.state

			# Set a prescribed burn at cell if it's unburnt
			if cell_state == CellStates.FUEL:
				fire.set_prescribed_fire_at_cell(cell_from_row_col)

		# Check if a wind time step has passed and print out the updated speed and direction
		if fire.curr_time_m - self.last_wind_update > fire.wind_vec.time_step:
			print("Wind updated")
			
			# Get the new wind conditions
			speed_m_s, dir_deg = fire.get_curr_wind_speed_dir()

			print(f"Wind speed: {speed_m_s} m/s")
			print(f"Wind direction: {dir_deg} deg")

			self.last_wind_update = fire.curr_time_m

		# Print the average fire coordinate every 100 iterations of the sim
		if fire.iters % 100 == 0:
			x_avg, y_avg = fire.get_avg_fire_coord()
			
			print(f"Avg. Fire Coordinate: ({x_avg} m, {y_avg} m)")


from setuptools import setup, find_packages

setup(
    name="embrs",
    version="0.0.13",
    author="Rui de Gouvea Pinto",
    author_email="rjdp3@gatech.edu",
    description="A fire simulation tool optimized for autonomous firefighting systems development",
    url="https://github.com/AREAL-GT/embrs",
    packages=find_packages(),
    install_requires=[
        "numpy>=1.0",
        "shapely>=2.0",
		"matplotlib>=3.0",
		"PyQt5>=5.0",
		"tqdm>=4.0",
		"pandas>=2.0",
		"rasterio>=1.0",
		"requests>=2.0",
		"pyproj>=3.0",
		"utm>=0.6",
		"scipy>=1.0"
    ],
    python_requires=">=3.8",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
	entry_points={
        "console_scripts": [
            "run_embrs_sim = embrs.main:main",
            "create_embrs_map = embrs.map_generator:main",
            "create_embrs_wind = embrs.wind_forecast_generator:main",
            "run_embrs_viz = embrs.visualization_tool:main"
        ],
	},
)



"""Module containing code that loads parameters from user input and runs the simulations
"""
from time import time
import signal
import importlib
import pickle
import json
import copy
import os
from typing import Tuple
from tqdm import tqdm
import numpy as np
from shapely.geometry import shape, LineString
from .fire_simulator.visualizer import Visualizer
from .fire_simulator.logger import Logger
from .fire_simulator.fire import FireSim
from .fire_simulator.wind import Wind
from .utilities.file_io import SimFolderSelector, LoaderWindow
from .utilities.fire_util import UtilFuncs
from .base_classes.control_base import ControlClass

def initialize(params: dict) -> Tuple[FireSim, bool, Visualizer]:
    """Initializes a fireSim object and a Visualizer(if selected) based on user inputs

    :param params: dictionary containing all user's inputs and preferenes, generated by
                   SimFolderSelector class.
    :type params: dict
    :return: tuple containing a FireSim object, a bool variable specifying if real-time 
             visualization was selected, and a Visualizer object. If real-time visualization not
             selected Visualizer will be None.
    :rtype: Tuple[FireSim, bool, Visualizer]
    """
    # Folder Setup
    input_folder = params['input']
    input_foldername = os.path.basename(input_folder)

    # Real-time viz
    viz_on = params['viz_on']

    # Map Setup
    cell_size = params['cell_size'] # meters
    time_step = params['t_step'] # seconds
    sim_duration = params['sim_time']

    # Load wind forecast and initialize wind vector
    wind_forecast = []

    if not params['zero_wind']:

        wind_file = params['wind']

        with open(wind_file, 'r') as file:
            wind_data = json.load(file)

        forecast_time_step = wind_data['time_step_min']

        for entry in wind_data['data']:
            wind_forecast.append((entry['speed_m_s'], entry['direction']))

    else:
        forecast_time_step = 1e10
        wind_forecast = [(0, 0)]

    wind_forecast = np.array(wind_forecast)

    # Initialize wind vector
    wind_vec = Wind(wind_forecast, forecast_time_step)

    # Read JSON map file
    with open(input_folder + '/' + input_foldername + '.json', 'r') as f:
        data = json.load(f)

    # Load data files
    topography_map = np.load(data['topography']['file'])
    fuel_map = np.load(data['fuel']['file'])
    ignition_dicts = data['initial_ignition']
    initial_ignition = [shape(d) for d in ignition_dicts]
    fire_breaks = [{'geometry': LineString(fb['geometry']['coordinates']),
                    'fuel_value': fb['fuel_value']} for fb in data['fire_breaks']]

    if 'roads' in data:
        with open(data['roads']['file'], 'rb') as f:
            roads = pickle.load(f)
    else:
        roads = None

    # Save resolution, width, and height into variables
    topography_resolution = data['topography']['resolution']
    fuel_resolution = data['fuel']['resolution']
    width_m = data['topography']['width_m']
    height_m = data['topography']['height_m']

    # Calculate simSize
    num_cols = int(np.floor(width_m/(np.sqrt(3)*cell_size)))
    num_rows = int(np.floor(height_m/(1.5*cell_size)))

    sim_size = (num_rows, num_cols)
    print("simSize: " + str(sim_size))

    # Initialize the simulator
    fire = FireSim(fuel_map, fuel_resolution, topography_map, topography_resolution, wind_vec,
                   roads, fire_breaks, time_step = time_step, initial_ignition = initial_ignition,
                   size = (width_m, height_m), cell_size=cell_size, duration_s = sim_duration)

    if viz_on:
        viz = Visualizer(fire)
    else:
        viz = None

    return fire, viz_on, viz

def run_sim(fire: FireSim, viz: Visualizer, progress_bar: tqdm, loader_window: LoaderWindow,
            viz_on: bool, user_code_path: str, user_code_class: str, i:int, num_runs:int):
    """Function that runs the fire simulation and user code. Runs until the sim duration is reached
    or the fire is extinguished.

    :param fire: FireSim object to be run
    :type fire: FireSim
    :param viz: Visualizer object used for real-time visualization
    :type viz: Visualizer
    :param progress_bar: tqdm progress bar for command line display
    :type progress_bar: tqdm
    :param loader_window: LoaderWindow object used for generating a tk window that displays 
                          progress
    :type loader_window: LoaderWindow
    :param viz_on: boolean variable specifying whether the sim should be displayed in real-time
                   or not
    :type viz_on: bool
    :param user_code_path: Path to the user's selected module to run in parallel with sim
    :type user_code_path: str
    :param user_code_class: Name of the user's selected class to run in parallel with sim
    :type user_code_class: str
    :param i: Current number of simulation running
    :type i: int
    :param num_runs: Number of total simulations to be run
    :type num_runs: int
    """
    start = time()

    time_since_last_update = 20
    last_update = 0

    if user_code_path == "":
        user = None
    else:
        # Strip the extension and path
        user_module_name = os.path.splitext(os.path.basename(user_code_path))[0]

        # Dynamically import the module
        user_module = importlib.import_module(user_module_name)

        # Access the class within the module
        UserCodeClass = getattr(user_module, user_code_class)

        # Check if the user's class is a subclass of the abstract base class
        if not issubclass(UserCodeClass, ControlClass):
            raise TypeError(f"{user_code_class} is not a subclass of ControlClass")

        user = UserCodeClass(fire)  # Construct the user object

    fire.logger.start_status_log()

    while not fire.finished:

        fire.iterate()

        if user:
            user.process_state(fire)

        progress_bar.update()
        loader_window.increment_progress()

        # ----------- Perform real-time plotting ------------- #
        if viz_on:
            time_since_last_update = fire.curr_time_s - last_update

            if time_since_last_update >= fire.display_frequency:
                viz.update_grid(fire)
                last_update = fire.curr_time_s

    iters_skipped = ((fire.sim_duration/fire.time_step) + 1) - fire.iters
    progress_bar.update(iters_skipped)
    loader_window.increment_progress(increment_value=iters_skipped)
    end = time()
    elapsed = end - start

    msg = (f"Simulation complete. Sim time: {UtilFuncs.get_time_str(fire.curr_time_s)}, "
        f"took {UtilFuncs.get_time_str(elapsed, True)} seconds to compute.")

    fire.logger.log_message(msg)

    loader_window.set_text(f"Saving log file for simulation ({i+1}/{num_runs})")
    fire.logger.finish(fire)

def handle_interrupt(logger, fire):
    """Function to handle user interrupts while a sim is running.
    """
    print("\nUser interrupt detected. Finalizing logging...")
    logger.finish(fire, True)

def sim_loop(params: dict):
    """Main function that gets user input from GUI and runs the specified simulation(s)
    """
    num_runs = params['num_runs']
    duration = params['sim_time']
    time_step = params['t_step']

    # Get user code info
    user_code_path = params['user_path']
    user_code_class = params['class_name']

    # Create loader window object
    loader_window = LoaderWindow('Fire sim', max_value=num_runs*(duration/time_step)+1)
    loader_window.set_text("Initializing simulations...")

    fire, viz_on, viz = initialize(params)

    data_logger = Logger(params['log'])
    data_logger.store_metadata(params, fire)
    data_logger.store_init_fire_obj(fire)

    print("Running simulations... ")
    print("")

    # Create command line progress bar
    progress_bar = tqdm(total=(num_runs*(fire.sim_duration/fire.time_step)),
                        desc='Total progress', position=1, leave=True)

    curr_fire = None

    for i in range(num_runs):
        if num_runs > 1:
            curr_fire = copy.deepcopy(fire)

        else:
            curr_fire = fire

        # Register the signal handler
        signal.signal(signal.SIGINT, lambda signum, frame: handle_interrupt(data_logger, curr_fire))

        curr_fire.logger = data_logger
        loader_window.set_text(f"Running simulation ({i+1}/{num_runs})")
        run_sim(curr_fire, viz, progress_bar, loader_window, viz_on, user_code_path,
                user_code_class, i, num_runs)

        if viz_on:
            viz.reset_figure(done = i==num_runs-1)

    progress_bar.close()
    loader_window.set_text(f"Finished running {num_runs} simulations(s)")
    loader_window.close()

    print("")
    print(f"Finished running {num_runs} simulation(s)")

def main():
    folder_selector = SimFolderSelector(sim_loop)
    folder_selector.run()

if __name__ == "__main__":
    main()



"""Logger for saving sim data and creating log files

.. autoclass:: Logger
    :members:
"""

import datetime
import pickle
import json
import sys
import os
import msgpack

from embrs.fire_simulator.fire import FireSim

class Logger:
    """Logger class used to generate log files from sim data

    :param log_folder: String representing the path to the folder where the logger instance will
                       write log files.
    :type log_folder: str
    """
    def __init__(self, log_folder: str):
        """Constructor method that saves input data and creates a logger instance.
        """
        self.time = 0
        self.log_ctr = 0
        self.cache = {}
        self.agent_cache = {}
        self.messages = []
        self.agent_data = False
        self._log_folder = log_folder
        self._session_folder = self.generate_session_folder()
        os.makedirs(self.session_folder, exist_ok=True)

        self.data = None
        self.metadata = None

    def generate_session_folder(self) -> str:
        """Generates the path for the current sim's log files based on current datetime

        :return: Session folder path string
        :rtype: str
        """
        date_time_str = datetime.datetime.now().strftime('%d-%b-%Y-%H-%M-%S')
        return os.path.join(self._log_folder, f"log_{date_time_str}")

    def store_init_fire_obj(self, fire_obj: FireSim):
        """Store the initial state of the fire in a .pkl within the log folder.
        
        This is done to speed up the post-processing of log data significantly.

        :param fire_obj: :class:`~fire_simulator.fire.FireSim` object just after being initialized.
        :type fire_obj: :class:`~fire_simulator.fire.FireSim`
        """
        fire = {
            'cell_size': fire_obj.cell_size,
            'grid_width': fire_obj.grid_width,
            'grid_height': fire_obj.grid_height,
            'time_step': fire_obj.time_step,
            'cell_dict': fire_obj.cell_dict,
            'cell_grid': fire_obj.cell_grid,
            'fire_breaks': fire_obj.fire_breaks,
            'topography_map': fire_obj.topography_map,
            'coarse_topography': fire_obj.coarse_topography,
            'fuel_map': fire_obj.fuel_map,
            'wind_vec': fire_obj.wind_vec
        }

        if fire_obj.roads is not None:
            fire['roads'] = fire_obj.roads


        filename = f"{self.session_folder}/init_fire_state.pkl"
        with open(filename, 'wb') as f:
            pickle.dump(fire, f)

    def add_to_cache(self, updated_cells: list, time: int):
        """Add the list of updated cells from the previous iteration of the sim to the logger's
        cache to be stored in a log file later on.

        :param updated_cells: list of :class:`~fire_simulator.cell.Cell` objects that were changed
                              in the last iteration in their log format. See
                              :func:`~fire_simulator.cell.Cell.to_log_format()` for more info.
        :type updated_cells: list
        :param time: sim time (in seconds) when these cells were updated and stored.
        :type time: int
        """
        self.cache[time] = updated_cells

    def add_to_agent_cache(self, agents: list, time: int):
        """Add list of agents' state from the previous iteration of the sim to the logger's agent
        cache to be stored in a log file later on.

        :param agents: list of :class:`~agent_base.AgentBase` instances in their log format see the
                       :class:`~agent_base.AgentBase`:func:`~agent_base.AgentBase.to_log_format()`
                       function for more info.
        :type agents: list
        :param time: sim time (in seconds) when agents' were stored.
        :type time: int
        """
        self.agent_data = True
        self.agent_cache[time] = agents

    def _dump_cache(self):
        """Dump the logger's cache in a .msgpack file in the specified file location to be used
        later in post-processing.
        """

        if not self.cache:
            return

        run_folder = f"{self.session_folder}/run_{self.log_ctr}"
        os.makedirs(run_folder, exist_ok=True)

        file_name = f"{run_folder}/log.msgpack"

        if self.agent_data:
            agent_filename = f"{run_folder}/agents.msgpack"

        try:
            with open(file_name, 'wb') as f:
                msgpack.pack(self.cache, f)

            if self.agent_data:
                with open(agent_filename, 'wb') as f:
                    msgpack.pack(self.agent_cache, f)

            self.log_message("Log data saved successfully!")

        except Exception as e:
            self.log_message(f"Exception {e} occurred while attempting to save data.")

    # ~~~ STATUS LOGGING ~~~ #
    def store_metadata(self, params: dict, fire: FireSim):
        """Store metadata for the status log describing the parameters used to run a simulation.
        
        Resulting metadata will be used in all status logs generated with a given simulation
        setup.

        :param params: dictionary containing input parameters for the simulation.
        :type params: dict
        :param fire: :class:`~fire_simulator.fire.FireSim` instance for which metadata is being
                     stored.
        :type fire: :class:`~fire_simulator.fire.FireSim`
        """
        # inputs
        cell_size = params['cell_size']
        time_step = params['t_step'] # seconds
        duration = params['sim_time']
        wind_file = params['wind']
        import_roads = fire.roads is not None

        # sim size
        rows = fire.shape[0]
        cols = fire.shape[1]
        total_cells = rows*cols
        width_m = fire.size[0]
        height_m = fire.size[1]

        # Load map file
        map_folder = params['input']
        foldername =  os.path.basename(map_folder)
        map_file_path = os.path.join(map_folder, foldername + ".json")
        with open(map_file_path, 'r') as f:
            map_data = json.load(f)

        user_code_path = params['user_path']
        user_code_class = params['class_name']

        metadata = {
                "inputs": {
                    "cell size": cell_size,
                    "time step (sec)": time_step,
                    "duration (sec)": duration,
                    "roads imported": import_roads
                },

                "sim size": {
                    "rows": rows,
                    "cols": cols,
                    "total cells": total_cells,
                    "width (m)": width_m,
                    "height (m)": height_m  
                },

                "wind forecast": {
                    "file location": wind_file
                },

                "imported_code": {
                    "imported module location": user_code_path,
                    "imported class name": user_code_class
                },

                "map": {
                    "map file location": map_file_path,
                    "map contents": map_data
                }
        }

        self.metadata = metadata

    def start_status_log(self):
        """Start a status log by loading the relevant metadata and creating a dictionary to store
        all relevant status data.
        """
        # loads metadata and creates a dictionary to store status data
        self.data = {}
        self.data["sim start"] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.data["metadata"] = self.metadata
        self.messages = []
        self.log_message("Initialization successful, simulation started.")

    def log_message(self, message: str):
        """Log a custom message in the current status log.

        :param message: Message to be logged
        :type message: str
        """
        # log a time-stamped message
        timestamp = self.create_timestamp()
        status = f"[{timestamp}]:{message} "

        self.messages.append(status)

    def _write_status_log(self, fire: FireSim, on_interrupt: bool = False):
        """Writes the status log after a simulation has been completed.
        
        :param fire: :class:`~fire_simulator.fire.FireSim` instance to write status log for.
        :type fire: :class:`~fire_simulator.fire.FireSim`
        """
        self.data["messages"] = self.messages

        if fire is not None:
            # log results of simulation run
            burning_cells = len(fire.curr_fires)
            burnt_cells = len(fire.burnt_cells)
            fire_extinguished = burning_cells == 0

            self.data["results"] = {
                "user interrupted": on_interrupt,
                "cells burnt": burnt_cells,
                "burnt area (m^2)": burnt_cells * fire.cell_dict[0].cell_area,
                "fire extinguished": fire_extinguished,
            }

            if not fire_extinguished:
                self.data["results"]["burning cells remaining"] = burning_cells
                self.data["results"]["burning area remaining (m^2)"] = burning_cells * fire.cell_dict[0].cell_area

        else:
            self.log_message("Unable to publish results due to early termination")
            self.data["results"] = None

        filename = f"{self.session_folder}/run_{self.log_ctr}/status_log.json"

        with open(filename, 'w') as f:
            json.dump(self.data, f, indent=4)

        self.log_ctr += 1

    def finish(self, fire: FireSim, on_interrupt=False):
        """Finish logging all data, write to a .msgpack log file and a .json status log.

        :param fire: :class:`~fire_simulator.fire.FireSim` instance to write logs for.
        :type fire: :class:`~fire_simulator.fire.FireSim`
        :param on_interrupt: set to `True` when writing final log file, if `True` program will
                             terminate when finished logging, defaults to `False`.
        :type on_interrupt: bool, optional
        """
        self._dump_cache()
        self._write_status_log(fire, on_interrupt)

        if on_interrupt:
            sys.exit(0)

    @property
    def session_folder(self) -> str:
        """Path to the folder the logger is writing to
        """
        return self._session_folder

    def create_timestamp(self) -> str:
        """Function that creates a timestamp string based on the current datetime and returns it.

        :return: Timestamp string representing the current date and time.
        :rtype: str
        """
        return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")



"""Demo class demonstrating how the use of the fire prediction tool during a fire simulation.
The prediction tool can be used to inform firefighting decision making.

To run this example code, start a fire sim and select this file as the "User Module"
"""

from base_classes.control_base import ControlClass
from fire_simulator.fire import FireSim
from tools.fire_predictor import FirePredictor

class PredictorCode(ControlClass):
    def __init__(self, fire:FireSim):
        self.prediction = False

    def process_state(self, fire: FireSim) -> None:
        if not self.prediction and fire.curr_time_h > 1:
            
            # Define time horizon to predict over
            time_horizon_hr = 2

            # Initialize the fire prediction tool
            predictor = FirePredictor(fire, time_horizon_hr, bias=1)

            # Run prediction
            future_fires = predictor.run_prediction()
            self.prediction = True

            # Print out the predicted fire locations for each time step
            for time_step in future_fires.keys():
                print(f"Time: {time_step}, fires: {future_fires[time_step]}")